#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2014 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#
try:
    import os
    import select
    import socket
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")


class clagdcmd(object):

    """
    Simple class that connects to the CLAG daemon's command socket, exported as 
    a UNIX_AF socket by clagd, allowing users to interact with clagd, sending   
    commands and getting results.  clagdcmd opens the socket file and issues    
    commands via clagdcmd.run().  The command output is read from the socket and
    returned to the caller.                                                     
    """

    version = "1.0"


    def __init__(self, socketname="/var/run/clagd.socket"):

        """
        Constructor:

        socketname - the name of the clagdd socket file.
        """

        if not os.access(socketname, os.F_OK):
            raise ValueError("socket %s does not exist" % socketname)
        elif not os.access(socketname, os.R_OK | os.W_OK):
            raise ValueError("missing read/write permissions for %s" % socketname)
        else:
            self.socketname = socketname

        self.socketobj = None
        self.__open__()


    def __del__(self):

        """Destructor: flush and close all files that we opened"""

        try:
            self.close()
        except:
            pass


    def __str__(self):

        """Return object state in human-readable form"""
        
        s = []
        s.append(repr(self))
        s.append('version: ' + self.version)
        s.append('socketname: ' + str(self.socketname))
        s.append('socketobj: ' + str(self.socketobj))
        return '\n'.join(s)


    def close(self):

        """Close the socket object"""

        if self.socketobj is not None:
            self.socketobj.shutdown(socket.SHUT_RDWR)
            self.socketobj.close()
            self.socketobj = None


    def __open__(self):

        """
        Connect to the a clagd socket. No-op if the socket is already open.
        """

        if self.socketobj is None:
            try:
                self.socketobj = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            except socket.error as msg:
                self.socketobj = None
                raise IOError("Unable to create a socket (%s)" % (msg,))

            try:
                self.socketobj.connect(self.socketname)
            except socket.error as msg:
                self.socketobj.close()
                self.socketobj = None
                raise IOError("Unable to connect to clagd command socket: %s (%s)" % (self.socketname, msg))
            self.buffer=''


    def run(self, cmd, callback=None, keepOpen=False):

        """
        Issue the command to the clagd and collect the return data.
        cmd must be a string. 
        """

        if type(cmd) is not str:
            raise TypeError("expecting string argument to clagdcmd.run(cmd)")
        cmd += chr(0)

        self.__open__()

        inputs = [self.socketobj]
        outputs = [self.socketobj]
        runResult = ""

        while inputs + outputs:
            (readable, writable, exceptional) = select.select(inputs, outputs, inputs + outputs)

            for readSock in readable:
                try:
                    cmdResult = readSock.recv(2048)
                except:
                    cmdResult = 0
                if cmdResult:
                    self.buffer += cmdResult
                    while "\n" in self.buffer and callback is not None:
                        cmdEnd = self.buffer.find("\n")
                        cmdStr = self.buffer[:cmdEnd]
                        self.buffer = self.buffer[cmdEnd+1:]
                        callback(cmdStr)
                    if chr(0) in self.buffer:
                        cmdEnd = self.buffer.find(chr(0))
                        runResult = self.buffer[:cmdEnd]
                        self.buffer = self.buffer[cmdEnd+1:]
                        if keepOpen:
                            return runResult
                        if readSock in inputs:
                            inputs.remove(readSock)
                else:
                    if readSock in inputs:
                        inputs.remove(readSock)
                        if readSock == self.socketobj:
                            self.close()

            for writeSock in writable:
                try:
                    numSent = writeSock.send(cmd)
                except:
                    numSent = len(cmd)
                    writeSock.close()
                cmd = cmd[numSent:]
                if not cmd and writeSock in outputs:
                    outputs.remove(writeSock)

            for errSock in exceptional:
                if errSock in inputs:
                    inputs.remove(errSock)
                if errSock in outputs:
                    outputs.remove(errSock)
                if errSock == self.socketobj:
                    self.close()

        return runResult

