#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2015 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    from kcache import KCache
    import os
    import threading
    import select
    import socket
    import rtnetlink as rtnl
    import iff
    import errno
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

class ClagNl(KCache):
    def __init__(self, Daemon, Parser, Log, Intf):
        self.Daemon = Daemon
        self.Log = Log
        self.Parser = Parser
        self.Intf = Intf

        # Data shared by all the threads
        self.keep_going = True

        self.familyFilters = { socket.AF_UNSPEC }

        # PID_MAX_LIMIT is 2^22 allowing 1024 sockets per-pid. We
        # start with 1 in the upper space (top 10 bits) instead of
        # 0 to avoid conflicts with netlink_autobind which always
        # attempts to bind with the pid (and on failure with 
        # negative values -4097, -4098, -4099 etc.)
        nl_port = os.getpid() | (1 << 22)
        # Setup initial iface cache
        KCache.__init__(self, nl_port, types=[rtnl.RTM_GETLINK], 
                        filters=self.filters(), notifiers=self.notifiers())

        # Start a thread to process nl cache updates - XXX - this is temporary
        # All threads that require access to the KCache will be collapsed to
        # avoid the need for event locking
        self.clagNlT = threading.Thread(None, self.ClagNlCacheT, "ClagNlCacheT")
        self.clagNlT.daemon = True
        self.clagNlT.start()

    def __del__(self):
        # Signal the update thread to wake up and exit
        self.SignalThreads()
        self.JoinThreads()

    def SignalThreads(self):
        self.keep_going = False
        # XXX find a better way to wake the thread
        try:
            self.sendto('', (self.pid, rtnl.RTMGRP_LINK))
        except socket.error as (e, string):
            self.log_debug("ClagNlSignalThreads: send err[%d]: %s" % \
                (e, string))

    def JoinThreads(self):
        self.clagNlT.join()
        #close the netlink socket
        socket.socket.close(self)

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_NETLINK, logStr)

    def ClagNlCacheT(self):
        try:
            self.ClagNlCache()
        except:
            self.Daemon.DumpTraceback()

    def CmpIface(self, old_iface, new_iface):
        return old_iface.ifname == new_iface.ifname and \
            old_iface.operstate == new_iface.operstate
        
    def AddIface(self, type, iface_info):
        ifindex = iface_info[0]
        old_iface = iface_info[1]
        new_iface = iface_info[2]

        # Is this a clag bond?
        if self.Parser.isClagBond(new_iface.ifname):
            if not (old_iface and self.CmpIface(old_iface, new_iface)):
                self.log_debug('AddIface: 0x%x' % ifindex)
                if old_iface:
                    self.log_debug('old_if(0x%x:%s): oper: %d' % (ifindex, old_iface.ifname, old_iface.operstate))
                self.log_debug('new_if(0x%x:%s): oper: %d' % (ifindex, new_iface.ifname, new_iface.operstate))

                # Currently we only process operstate changes
                if old_iface and old_iface.operstate == new_iface.operstate:
                    return
                if new_iface.operstate == iff.IF_OPER_DORMANT:
                    self.Log.log_debug(self.Log._LOG_DEBUG_FSM, '%s state dormant' % (new_iface.ifname))
                    self.Intf.SetDormantBondOperState(new_iface.ifname)
        # Is this a VxLAN interface?
        elif new_iface.linkinfo and new_iface.linkinfo.get(rtnl.IFLA_INFO_KIND) == 'vxlan':
            data = new_iface.linkinfo.get(rtnl.IFLA_INFO_DATA)
            if data:
                brportst = new_iface.protinfo.get(rtnl.IFLA_BRPORT_STATE, 0)
                vni     = data.get(rtnl.IFLA_VXLAN_ID)
                localip = str(data.get(rtnl.IFLA_VXLAN_LOCAL))
                opstate = iff.operstate_str.get(new_iface.operstate, "unknown")
                if new_iface.operstate == iff.IF_OPER_DOWN and not new_iface.flags & iff.IFF_UP:
                    opstate = "admin-down"
                self.log_debug('*AddIface: ifindex 0x%x, ifname %s, operstate %s, local %s, vid %d, brportst %d' % (ifindex, new_iface.ifname, opstate, localip, vni, brportst))
                self.Intf.AddVxLanInterface(new_iface.ifname, vni, localip, opstate, brportst)
        
    def DelIface(self, type, iface_info):
        ifindex = iface_info[0]
        iface = iface_info[1]
        self.log_debug('DelIface: 0x%x' % ifindex)
        self.Intf.DelVxLanInterface(iface.ifname)

    def SetClearFilter(self, enable, disable):
        self.familyFilters = (self.familyFilters | enable) - disable

    def Filter(self, type, ifi):
        self.log_debug('*Filter: family %d, type %d, index %d, flags 0x%x, change 0x%x, filter %s' % (ifi.ifi_family, ifi.ifi_type, ifi.ifi_index, ifi.ifi_flags, ifi.ifi_change, str(ifi.ifi_family not in self.familyFilters)))
        return ifi.ifi_family not in self.familyFilters

    def GetOperState(self, name):
        iface = self.iface_by_name(name)
        if iface and iface.operstate in iff.operstate_str:
            return iff.operstate_str[iface.operstate]
        return "unknown"

    def GetKCacheLinkCopy(self):
        return self.iface_db_copy()

    def filters(self):
        return {
            rtnl.RTM_NEWLINK: self.Filter,
            rtnl.RTM_DELLINK: self.Filter,
        }

    def notifiers(self):
        return {
            rtnl.RTM_NEWLINK: self.AddIface,
            rtnl.RTM_DELLINK: self.DelIface,
        }

    def ClagNlCache(self):
        epoll = select.epoll()
        epoll.register(self.fileno(), select.EPOLLIN)
        #self.log_iface_cache()
        while self.keep_going:
            try:
                events = epoll.poll()
            except IOError, e:
                if e.errno != errno.EINTR:
                    raise
                events = []
            if self.keep_going:
                for fileno, event in events:
                    if self.keep_going and fileno == self.fileno():
                        self.process()

    def log_iface_cache(self):
        self.log_debug('log_iface_cache')
        for ifindex, iface in self.ifaces.items():
            self.log_debug('if(0x%x:%s): oper: %d, linkinfo: %s, protinfo: %s' %(ifindex, iface.ifname, iface.operstate, str(iface.linkinfo), str(iface.protinfo)))
