#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2014 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import copy
    import gc
    import os
    import platform
    import select
    import socket
    import struct
    import subprocess
    import sys
    import threading
    import traceback
    import json
    import iff
    import clag.util as util
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

yappiAvail = True
try:
    import yappi
except ImportError, e:
    yappiAvail = False

class cmdsrv:
    """
    This class handles everything required to implement the command server
    interface for clagd. This interface allows other processes to communicate
    with clagd to get status and alter configuration. clagctl is an example
    of a process which connects to this interface.
    """

    _CmdSocketFile = "/var/run/clagd.socket"

    #   XXX: This is a hack. Python socket module doesn't define the SO_PEERCRED
    #   value. And this value is different on different CPU architectures. So on
    #   PowerPC platforms the value is 21 and on all other platforms the value is 17.
    _SOCKET_PEERCRED = 21 if platform.machine() == "ppc" else 17
    _JSON_SORT_KEYS = True
    _JSON_INDENT = 4


    def __init__(self, Daemon, Intf, Parser, Log, LacpSync, VxLanSync, FdbSync, MdbSync, VlanSync, HealthCheck):
        """
        Constructor: Initialize data structures, threads, etc.
        """

        #
        #   Save the parameters passed in
        #
        self.Daemon     = Daemon
        self.Intf       = Intf
        self.Parser     = Parser
        self.Log        = Log
        self.LacpSync   = LacpSync
        self.VxLanSync  = VxLanSync
        self.FdbSync    = FdbSync
        self.MdbSync    = MdbSync
        self.VlanSync   = VlanSync
        self.HealthCheck = HealthCheck


        #
        #   CLAGD command interface sockets. The cmdSrvSock is a socket to which clients
        #   can connect to open up communication with the clagd daemon. The cmdOutputBufs
        #   is a dictionary keyed by the client socket, with the string of data to write
        #   to the socket as the value. Only sockets which have data waiting to be output 
        #   are in this dictionary. The cmdGetAll set is a list of sockets which are 
        #   currently in the "getall" state, monitoring the status of CLAGD.
        #
        self.cmdSrvSock         = None
        self.cmdOutputLock      = threading.Lock()
        self.cmdOutputBufs      = {}
        self.cmdGetAll          = set()

        #
        #   Start a thread to handle requests from other tools
        #
        self.keep_going         = True
        self.commandThread = threading.Thread(None, self.CmdRecvT, "CmdRecv")
        self.commandThread.daemon = True
        self.commandThread.start()

    def CmdRecvT(self):
        try:
            self.CmdRecv()
        except:
            self.Daemon.DumpTraceback()


    def __del__(self):
        """
        Destructor:
        """
        #
        #   Signal the command thread to stop and wait for it to die
        #
        self.SignalThreads()
        self.JoinThreads()

    def SignalThreads(self):
        self.keep_going = False

    def JoinThreads(self):
        if self.cmdSrvSock is not None:
            cmdSrvSock = self.cmdSrvSock
            self.cmdSrvSock = None
            cmdSrvSock.shutdown(socket.SHUT_RDWR)
        self.commandThread.join()

    def CmdRecv(self):
        '''
        This function is an independent thread of the script. It creates a socket,
        cmdSrvSock, and accepts connections on that socket. It receives the commands
        sent through those connections and replies with the proper command responses.
        '''
        self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Beginning execution of the thread CmdRecv")

        # Create a socket upon which we will accept connections
        try:
            self.cmdSrvSock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        except socket.error as msg:
            self.cmdSrvSock = None
            self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Unable to create a Unix stream socket (%s)" % (msg,))
        else:
            try:
                os.remove(cmdsrv._CmdSocketFile)
            except OSError:
                pass

            try:
                self.cmdSrvSock.bind(cmdsrv._CmdSocketFile)
                os.chmod(cmdsrv._CmdSocketFile, 0777)
                self.cmdSrvSock.listen(5)
                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Server socket to %s was created" % (cmdsrv._CmdSocketFile))
            except socket.error as msg:
                self.cmdSrvSock.close()
                os.remove(cmdsrv._CmdSocketFile)
                self.cmdSrvSock = None
                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Unable to bind socket to %s (%s)" % (cmdsrv._CmdSocketFile, msg))

        inputBufs = {}
        inputs  = [ self.cmdSrvSock ]
        while (inputs and self.cmdSrvSock and self.keep_going) or self.cmdOutputBufs:

            self.cmdOutputLock.acquire()
            cmdOutputSocks = self.cmdOutputBufs.keys()
            self.cmdOutputLock.release()

            # Wait until there is a socket to read, write, or an error
            try:
                (readable, writable, exceptional) = select.select(inputs, cmdOutputSocks, inputs + cmdOutputSocks, self.Parser.args.cmdConnect)
            except select.error as e:
                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "A select.error occurred")
                (readable, writable, exceptional) = ([],[],[])
            except socket.error:
                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "A socket.error occurred")
                (readable, writable, exceptional) = ([],[],[])

            if self.keep_going:
                for readSock in readable:
                    if readSock is self.cmdSrvSock:
                        # Accept a new socket connection
                        self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Accepting a new cmd connection")
                        (newSock, clientAddr) = readSock.accept()
                        inputs.append(newSock)
                        inputBufs[newSock] = ""
                    else:
                        # Read data from this socket
                        self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Reading data from the command socket")
                        try:
                            inputBuf = readSock.recv(2048)
                        except:
                            inputBuf = 0
                        if inputBuf:
                            # Add the data read to this socket's receive buffer
                            inputBufs[readSock] += inputBuf
                            while chr(0) in inputBufs[readSock]:
                                inEnd = inputBufs[readSock].find(chr(0))
                                inCmd = inputBufs[readSock][:inEnd]
                                inputBufs[readSock] = inputBufs[readSock][inEnd+1:]
                                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "    %s" % (inCmd,))
                                self.HandleCommand(inCmd, readSock)
                        else:
                            # If no data was available, the socket has been closed, end of message.
                            if readSock in inputs:
                                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Closing a socket because there's nothing to read")
                                del inputBufs[readSock]
                                inputs.remove(readSock)
                                readSock.close()

            for writeSock in writable:
                self.cmdOutputLock.acquire()
                writeStr = self.cmdOutputBufs.get(writeSock)
                self.cmdOutputLock.release()
                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Sending data to a socket: %s" % (writeStr,))
                if writeStr is not None:
                    try:
                        numSent = writeSock.send(writeStr)
                    except:
                        self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Error sending data to the socket. Closing socket.")
                        numSent = len(writeStr)
                        writeSock.close()
                    self.cmdOutputLock.acquire()
                    self.cmdOutputBufs[writeSock] = writeStr[numSent:]
                    if not self.cmdOutputBufs[writeSock]:
                        self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "No more data to output right now on the socket")
                        del self.cmdOutputBufs[writeSock]
                    self.cmdOutputLock.release()

            for errSock in exceptional:
                self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Closing a socket because of an error")
                if errSock in inputs:
                    if errSock in inputBufs:
                        del inputBufs[errSock]
                    inputs.remove(errSock)
                self.cmdOutputLock.acquire()
                if errSock in self.cmdOutputBufs:
                    del self.cmdOutputBufs[errSock]
                self.cmdOutputLock.release()
                errSock.close()
                if errSock is self.cmdSrvSock:
                    self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "   The socket was the server socket")
                    self.cmdSrvSock = None

            # Check to see if any output sockets have closed
            self.cmdOutputLock.acquire()
            cmdOutputSocks = self.cmdOutputBufs.keys()
            self.cmdOutputLock.release()

            for cmdOutputSock in cmdOutputSocks:
                try:
                    cmdOutputSock.fileno()
                except socket.error:
                    self.cmdOutputLock.acquire()
                    del self.cmdOutputBufs[cmdOutputSock]
                    self.cmdOutputLock.release()

        self.Log.log_debug(self.Log._LOG_DEBUG_CMD_SERVER, "Finished execution of the thread CmdRecv")


    #-------------------------------------------------------------------------------
    #
    #   Command Interface Functions - These functions implement the command interface
    #   to the clagd daemon.
    #
    #-------------------------------------------------------------------------------

    def HandleCmdEcho(self, cmdStr, ourSock, cmdOutputBufs):
        self.cmdOutputBufs[ourSock] += cmdStr[5:]
        return chr(0)

    def HandleCmdSummary(self, cmdStr, ourSock, cmdOutputBufs):
        status = {}
        status['peerAlive'] = self.Intf.isPeerAlive()
        if self.Intf.isPeerAlive():
            status['peerPriority'] = self.Parser.peerId[0]
            status['peerId'] = self.Parser.peerId[1]
            status['peerRole'] = self.Parser.peerRole
        else:
            status['peerPriority'] = self.Parser.default_priority
            status['peerId'] = "00:00:00:00:00:00"
            status['peerRole'] = self.Parser.default_role

        status['ourPriority'] = self.Parser.clagId[0]
        status['ourId'] = self.Parser.clagId[1]
        status['ourRole'] = self.Parser.clagRole

        status['sysMac'] = self.Parser.args.sysMac
        status['backupIp'] = self.Parser.args.backupIp
        if self.Parser.args.backupVrf:
            status['backupVrf'] = self.Parser.args.backupVrf
        status['backupActive'] = self.HealthCheck.backupActive
        if self.Parser.args.vxlanAnycast:
            status['vxlanAnycast'] = self.Parser.args.vxlanAnycast

        status['peerIf'] = self.Parser.args.peerIf
        status['peerIp'] = self.Parser.args.peerIp

        self.cmdOutputBufs[ourSock] += json.dumps(status, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return chr(0)

    def HandleCmdBackupInfo(self, cmdStr, ourSock, cmdOutputBufs):
        backup = {}
        backup['backupIp'] = self.Parser.args.backupIp
        if self.Parser.args.backupVrf:
            backup['backupVrf'] = self.Parser.args.backupVrf
        backup['backupActive'] = self.HealthCheck.backupActive
        backup['backupRole'] = self.HealthCheck.backupRole
        if self.HealthCheck.backupRoleAvailable:
            backup['peerPriority'] = self.HealthCheck.peerId[0]
            backup['peerId'] = self.HealthCheck.peerId[1]
            backup['peerRole'] = self.HealthCheck.peerRole
        else:
            backup['peerPriority'] = self.Parser.default_priority
            backup['peerId'] = "00:00:00:00:00:00"
            backup['peerRole'] = self.Parser.default_role

        self.cmdOutputBufs[ourSock] += json.dumps(backup, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return chr(0)

    def HandleCmdParams(self, cmdStr, ourSock, cmdOutputBufs):
        '''
        Implements the "params" command. Fill up a string with all of the interesting
        parameters of the clagd program.
        '''
        params = {}
        params['clagVersion'] = self.Parser._ClagVersion
        params['clagDataVersion'] = self.Parser._ClagDataVersion
        params['clagCmdVersion'] = self.Parser._ClagCmdVersion
        params['peerIp'] = self.Parser.args.peerIp
        params['peerIf'] = self.Parser.args.peerIf
        params['sysMac'] = self.Parser.args.sysMac
        params['lacpPoll'] = self.Parser.getCfgLacpPoll()
        params['currLacpPoll'] = self.Parser.getCurrLacpPoll()
        params['peerConnect'] = self.Parser.args.peerConnect
        params['cmdConnect'] = self.Parser.args.cmdConnect
        params['peerLinkPoll'] = self.Parser.getPeerLinkPoll()
        params['switchdReadyTimeout'] = self.Parser.args.switchdReadyTimeout
        params['reloadTimer'] = self.Parser.args.reloadTimer
        params['periodicRun'] = self.Parser.getPeriodicRun()
        params['priority'] = self.Parser.args.priority
        params['quiet'] = self.Log.getQuiet()
        params['debug'] = self.Log.getDebug()
        params['verbose'] = self.Log.getVerbose()
        params['log'] = self.Log.getHandlerStr()
        params['vm'] = self.Parser.args.vm
        params['peerPort'] = self.Parser.args.peerPort
        params['peerTimeout'] = self.Parser.args.peerTimeout
        params['sendTimeout'] = self.Parser.args.sendTimeout
        params['sendBufSize'] = self.Parser.args.sendBufSize
        params['forceDynamic'] = self.Parser.args.forceDynamic
        params['dormantDisable'] = self.Parser.args.dormantDisable
        params['backupIp'] = self.Parser.args.backupIp
        if self.Parser.args.backupVrf:
            params['backupVrf'] = self.Parser.args.backupVrf
        params['vxlanAnycast'] = self.Parser.args.vxlanAnycast
        params['backupPort'] = self.Parser.args.backupPort
        params['cmdLine'] = sys.argv

        self.cmdOutputBufs[ourSock] += json.dumps(params, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return chr(0)


    def HandleCmdBondSetClagId(self, cmdStr, ourSock, cmdOutputBufs):
        cmdList = cmdStr.split()
        try:
            bond = cmdList[1]
            clagId = int(cmdList[2])
        except:
            cmdOutputBufs[ourSock] += "ERROR SetClagId command requires <bond> <0-%d>: %s" % (self.Parser._CLAG_ID_MAX, cmdStr,)
            return chr(0)
        cmdOutputBufs[ourSock] += self.Parser.setBondClagId(bond, clagId)
        return chr(0)


    def HandleCmdReloadDone(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += self.Parser.setReloadDone()
        return chr(0)


    def HandleCmdSetBackupIp(self, cmdStr, ourSock, cmdOutputBufs):
        cmdList = cmdStr.split()
        try:
            address = cmdList[1]
        except:
            address = None
        cmdOutputBufs[ourSock] += self.Parser.setBackupIp(address)
        return chr(0)


    def HandleCmdDebug(self, cmdStr, ourSock, cmdOutputBufs):
        cmdList = cmdStr.split()
        try:
            self.Log.setDebug(int(cmdList[1],0))
        except:
            cmdOutputBufs[ourSock] += "ERROR Debug command requires an integer argument: %s" % (cmdStr,)
        return chr(0)


    def HandleCmdVerbose(self, cmdStr, ourSock, cmdOutputBufs):
        cmdList = cmdStr.split()
        try:
            self.Log.setVerbose(cmdList[1].lower() in ["yes", "true", "t", "1"])
        except:
            cmdOutputBufs[ourSock] += "ERROR SetVerbose command requires a True/False argument: %s" % (cmdStr,)
        return chr(0)


    def HandleCmdQuiet(self, cmdStr, ourSock, cmdOutputBufs):
        cmdList = cmdStr.split()
        try:
            self.Log.setQuiet(cmdList[1].lower() in ["yes", "true", "t", "1"])
        except:
            cmdOutputBufs[ourSock] += "ERROR SetQuiet command requires a True/False argument: %s" % (cmdStr,)
        return chr(0)


    def HandleCmdLogfile(self, cmdStr, ourSock, cmdOutputBufs):
        cmdList = cmdStr.split()
        try:
            self.Log.setHandler(cmdList[1])
        except:
            cmdOutputBufs[ourSock] += "ERROR SetLogFile command requires the name of a log file: %s" % (cmdStr,)
        return chr(0)


    def HandleCmdLacpPoll(self, cmdStr, ourSock, cmdOutputBufs):
        cmdList = cmdStr.split()
        try:
            self.Parser.args.lacpPoll = int(cmdList[1],0)
        except:
            cmdOutputBufs[ourSock] += "ERROR SetLacpPoll command requires an integer argument: %s" % (cmdStr,)
        return chr(0)


    def HandleCmdPeerTimeout(self, cmdStr, ourSock, cmdOutputBufs):
        cmdList = cmdStr.split()
        try:
            timeout = int(cmdList[1],0)
            if timeout == 0:
                timeout = None
            self.Parser.setPeerTimeout(timeout)
        except:
            cmdOutputBufs[ourSock] += "ERROR SetPeerTimeout command requires an integer argument: %s" % (cmdStr,)
        return chr(0)


    def HandleCmdSendTimeout(self, cmdStr, ourSock, cmdOutputBufs):
        cmdList = cmdStr.split()
        try:
            self.Parser.args.sendTimeout = int(cmdList[1],0)
            self.Parser.CloseClientSock()
        except:
            cmdOutputBufs[ourSock] += "ERROR SetSendTimeout command requires an integer argument: %s" % (cmdStr,)
        return chr(0)


    def HandleCmdSendBufSize(self, cmdStr, ourSock, cmdOutputBufs):
        cmdList = cmdStr.split()
        try:
            self.Parser.args.sendBufSize = int(cmdList[1],0)
            self.Parser.CloseClientSock()
        except:
            cmdOutputBufs[ourSock] += "ERROR SetSendBufSize command requires an integer argument: %s" % (cmdStr,)
        return chr(0)


    def HandleCmdPriority(self, cmdStr, ourSock, cmdOutputBufs):
        global clagId
        cmdList = cmdStr.split()
        try:
            newPriority = int(cmdList[1],0)
            if 0 <= newPriority <= 65535:
                self.Parser.clagId[0] = newPriority
            else:
                raise ValueError
        except ValueError:
            cmdOutputBufs[ourSock] += "ERROR SetPriority command requires an integer argument, 0-65535: %s" % (cmdStr,)
        return chr(0)


    def HandleCmdPeerLinkPoll(self, cmdStr, ourSock, cmdOutputBufs):
        cmdList = cmdStr.split()
        try:
            self.Parser.setPeerLinkPoll(int(cmdList[1],0))
        except:
            cmdOutputBufs[ourSock] += "ERROR SetPeerLinkPoll command requires an integer argument: %s" % (cmdStr,)
        return chr(0)


    def HandleCmdDualIfs(self, cmdStr, ourSock, cmdOutputBufs):
        dualIfDb = {}
        dualIfs = self.Intf.GetDualIfs()
        for dualIf in dualIfs:
            dualIfInfo={}
            if dualIfs[dualIf]:
                dualIfInfo['peerIf'] = dualIfs[dualIf]
                dualIfInfo['status'] = "dual"
            else:
                dualIfInfo['status'] = "single"
            dualIfInfo['clagId'] = self.Parser.GetBondClagId(dualIf)
            dualIfDb[dualIf] = dualIfInfo
        self.cmdOutputBufs[ourSock] += json.dumps(dualIfDb, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return chr(0)



    def HandleCmdPeerAlive(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += str(self.Intf.isPeerAlive())
        return chr(0)


    def HandleCmdClagId(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += " ".join([str(self.Parser.clagId[0]), str(self.Parser.clagId[1])])
        return chr(0)


    def HandleCmdClagRole(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += str(self.Parser.clagRole)
        return chr(0)


    def HandleCmdBackupPeerId(self, cmdStr, ourSock, cmdOutputBufs):
        if self.HealthCheck.backupRoleAvailable:
            cmdOutputBufs[ourSock] += " ".join([str(self.HealthCheck.peerId[0]), str(self.HealthCheck.peerId[1])])
        else:
            cmdOutputBufs[ourSock] += " ".join("-")
        return chr(0)


    def HandleCmdBackupPeerRole(self, cmdStr, ourSock, cmdOutputBufs):
        if self.HealthCheck.backupRoleAvailable:
            cmdOutputBufs[ourSock] += str(self.HealthCheck.peerRole)
        else:
            cmdOutputBufs[ourSock] += " ".join("-")
        return chr(0)

    def HandleCmdPeerId(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += " ".join([str(self.Parser.peerId[0]), str(self.Parser.peerId[1])])
        return chr(0)


    def HandleCmdPeerRole(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += str(self.Parser.peerRole)
        return chr(0)


    def HandleCmdPeerIf(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += str(self.Parser.args.peerIf)
        return chr(0)


    def HandleCmdPeerIp(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += str(self.Parser.args.peerIp)
        return chr(0)


    def HandleCmdGetBackupIp(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += str(self.Parser.args.backupIp) if self.Parser.args.backupIp else '-'
        return chr(0)


    def HandleCmdBackupActive(self, cmdStr, ourSock, cmdOutputBufs):
        if self.HealthCheck.backupActive:
            stateStr = 'active'
        else:
            stateStr = 'inactive'
        cmdOutputBufs[ourSock] += stateStr
        return chr(0)


    def HandleCmdBackupRole(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += 'primary' if self.HealthCheck.backupRole == self.Parser._ROLE_PRIMARY else 'secondary'
        return chr(0)


    def HandleCmdSysMac(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += str(self.Parser.args.sysMac)
        return chr(0)


    def HandleCmdOurLacpDB(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += self.LacpSync.EncodedOurStr()
        return chr(0)


    def HandleCmdPeerLacpDB(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += self.LacpSync.EncodedPeerStr()
        return chr(0)


    def HandleCmdPeerLacpRate(self, cmdStr, ourSock, cmdOutputBufs):
        self.Parser.peerLacpRateLock.acquire()
        cmdOutputBufs[ourSock] += str(self.Parser.peerLacpRate)
        self.Parser.peerLacpRateLock.release()
        return chr(0)


    def HandleCmdOurMacDB(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += self.FdbSync.EncodedOurStr()
        return chr(0)


    def HandleCmdGetQInfo(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += util.queueGetInfo()
        return chr(0)


    def HandleCmdClagBondDB(self, cmdStr, ourSock, cmdOutputBufs):
        self.HandleCmdClagIntfDB(cmdStr, ourSock, cmdOutputBufs, True)
        return chr(0)


    def HandleCmdClagIntfDB(self, cmdStr, ourSock, cmdOutputBufs, justBonds=False):
        intfDb = {}
        dualIfs = self.Intf.GetDualIfs()
        if justBonds:
            clagIntfs = self.Parser.GetClagBondDB()
        else:
            clagIntfs = self.Parser.GetClagIntfDB()
        for intf in clagIntfs.values():
            intfInfo = {}
            intfInfo['clagId'] = intf.clagId
            peerIf = dualIfs.get(intf.intfName)
            if peerIf:
                intfInfo['peerIf'] = peerIf
                intfInfo['status'] = "dual"
            else:
                intfInfo['status'] = "single"
            conflicts = self.Parser.GetClagIntfConflictStr(intf.intfName)
            if conflicts:
                intfInfo['conflicts'] = conflicts
            protoDownReasons = intf.GetProtoDownReasons()
            if protoDownReasons:
                intfInfo['protoDown'] = protoDownReasons
            intfDb[intf.intfName]=intfInfo
        self.cmdOutputBufs[ourSock] += json.dumps(intfDb, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return chr(0)


    def HandleCmdNlLinkCache(self, cmdStr, ourSock, cmdOutputBufs):
        linkCache = self.Intf.GetKCacheLinkCopy()
        linkCacheDb = {}
        if linkCache:
            for ifIndex, iface in linkCache.items():
                linkInfo = {}
                linkInfo['flags'] = iface.flags
                linkInfo['operState'] = iff.operstate_str[iface.operstate]
                linkCacheDb[iface.ifname] = linkInfo
        self.cmdOutputBufs[ourSock] += json.dumps(linkCacheDb, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        return chr(0)

    def DumpThreadStacks(self):
        id2name = dict([(th.ident, th.name) for th in threading.enumerate()])
        self.Log.log("DumpThreadStacks - start\n")
        for threadId, stack in sys._current_frames().items():
            self.Log.log("\nthread: %s (%d)\n" % (id2name.get(threadId,""), threadId))
            for filename, lineno, name, line in traceback.extract_stack(stack):
                self.Log.log('  file: %s, line %d, in %s\n' % (filename, lineno, name))
            self.Log.log("\n")
        self.Log.log("DumpThreadStacks - end\n")

    def HandleCmdIntInfo(self, cmdStr, ourSock, cmdOutputBufs):
        intInfo = {}
        intInfo['syncDoneFromPeer'] = self.Parser.syncDoneFromPeer
        intInfo['peerDataVersion'] = self.Parser.peerDataVersion
        intInfo['configReloadDone'] = self.Parser.configReloadDone
        self.cmdOutputBufs[ourSock] += json.dumps(intInfo, sort_keys=self._JSON_SORT_KEYS, indent=self._JSON_INDENT)
        self.DumpThreadStacks()
        return chr(0)


    def HandleCmdPeerMacDB(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += self.FdbSync.EncodedPeerStr()
        return chr(0)


    def HandleCmdOurMdbDB(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += self.MdbSync.EncodedOurStr()
        return chr(0)


    def HandleCmdPeerMdbDB(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += self.MdbSync.EncodedPeerStr()
        return chr(0)

    def HandleCmdOurRportDB(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += self.MdbSync.EncodedOurRportStr()
        return chr(0)


    def HandleCmdPeerRportDB(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += self.MdbSync.EncodedPeerRportStr()
        return chr(0)


    def HandleCmdPeerVlanDB(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += self.VlanSync.EncodedPeerStr()
        return chr(0)


    def HandleCmdOurVlanDB(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += self.VlanSync.EncodedOurStr()
        return chr(0)


    def HandleCmdLogMsg(self, cmdStr, ourSock, cmdOutputBufs):
        self.Log.log(cmdStr[7:])
        return chr(0)


    def HandleCmdGarbage(self, cmdStr, ourSock, cmdOutputBufs):
        collected = gc.collect()
        cmdOutputBufs[ourSock] += str(collected)
        return chr(0)


    def HandleCmdGetAll(self, cmdStr, ourSock, cmdOutputBufs):
        termChar = chr(0)
        if ourSock not in self.cmdGetAll:
            cmdOutputBufs[ourSock] += self.HandleMstpState()
            self.cmdGetAll = self.cmdGetAll | set([ourSock])
            termChar = ""
        else:
            cmdOutputBufs[ourSock] += "ERROR The GetAll command has already been enabled on this socket."
        return termChar


    def HandleCmdDumpMstp(self, cmdStr, ourSock, cmdOutputBufs):
        cmdOutputBufs[ourSock] += self.HandleMstpState()
        return chr(0)


    def HandleCmdNoGetAll(self, cmdStr, ourSock, cmdOutputBufs):
        if ourSock in self.cmdGetAll:
            self.cmdGetAll = self.cmdGetAll - set([ourSock])
            cmdOutputBufs[ourSock] += chr(0)
        return chr(0)


    def HandleMstpState(self):
        '''
        Dumps the current state of clagd (as far as mstpd is concerned) at the
        beginning of a getall command.
        '''
        paramStr = ""
        peerLink = self.Parser.args.peerIf
        if self.Intf.isSubIfName(peerLink):
            masters = self.Intf.GetLogicalMastersOfSubIf(peerLink)
            if len(masters):
                peerLink = masters[0]
        paramStr += "add peer-link = %s\n" % (peerLink,)
        paramStr += "add clag-state = %s\n" % ("up" if self.Intf.isPeerAlive() else "down",)
        paramStr += "add system-mac = %s\n" % (self.Parser.args.sysMac,)
        paramStr += "add clag-role = %s\n" % (str(self.Parser.clagRole),)

        myDualIfs = self.Intf.GetDualIfs()
        myPeerLacpDB = self.LacpSync.GetPeerLacpDB()

        for dualIf in myDualIfs:
            peerData = myPeerLacpDB.get(myDualIfs[dualIf])
            if peerData is not None:
                paramStr += "add mac-" + str(peerData[0]) + " = " + dualIf + "," + str(peerData[1]) + "\n"

        return paramStr


    def HandleStartProfile(self, cmdStr, ourSock, cmdOutputBufs):
        if yappiAvail:
            yappi.clear_stats()
            yappi.set_context_name_callback(lambda: threading.current_thread().name)
            yappi.start()
        else:
            cmdOutputBufs[ourSock] += "ERROR Cannot start profiling. Yappi is not installed."
        return chr(0)


    def HandleStopProfile(self, cmdStr, ourSock, cmdOutputBufs):
        if yappiAvail:
            yappi.stop()
        else:
            cmdOutputBufs[ourSock] += "ERROR Cannot stop profiling. Yappi is not installed."
        return chr(0)


    def HandleDumpProfile(self, cmdStr, ourSock, cmdOutputBufs):
        if yappiAvail:
            self.Log.log("%-40s %9s %12s %12s %12s" % ("Function", "ncall", "tsub", "ttot", "tavg"))
            self.Log.log("%-40s %9s %12s %12s %12s" % ("--------", "-----", "----", "----", "----"))
            for stat in yappi.get_func_stats().sort('totaltime'):
                self.Log.log("%-40s %9d %12.6f %12.6f %12.6f" % (stat[0][-40:], stat[3], stat[7], stat[6], stat[11]))

            self.Log.log("%-20s %12s %12s" % ("Thread", "ttot", "schedcnt"))
            self.Log.log("%-20s %12s %12s" % ("------", "----", "--------"))
            for stat in yappi.get_thread_stats().sort('totaltime'):
                self.Log.log("%-20s %12.6f %12d" % (stat[0][-20:], stat[3], stat[4]))
        else:
            cmdOutputBufs[ourSock] += "ERROR Cannot dump profiling. Yappi is not installed."
        return chr(0)


    def CmdGetAllOutput(self, outStr):
        '''
        This function is called by various threads when certain events happen to
        send output to sockets which are monitoring the state of CLAGD.
        '''
        self.cmdOutputLock.acquire()
        for ourSock in self.cmdGetAll:
            if ourSock not in self.cmdOutputBufs:
                self.cmdOutputBufs[ourSock] = ""
            self.cmdOutputBufs[ourSock] += outStr + "\n"
        self.cmdOutputLock.release()


    def HandleCommand(self, cmdStr, ourSock):
        '''
        This function is called from the CmdRecv thread when a command has been
        received. The command is parsed, executed,  and the output data, if any, 
        is added to the output buffer.
        '''

        clagCmds = {
        #   Command              Function                    Priviledged
            "echo"            : (self.HandleCmdEcho,         False),
            "params"          : (self.HandleCmdParams,       False),
            "setclagid"       : (self.HandleCmdBondSetClagId,True ),
            "setbackupip"     : (self.HandleCmdSetBackupIp,  True ),
            "reloaddone"      : (self.HandleCmdReloadDone,   True ),
            "debug"           : (self.HandleCmdDebug,        True ),
            "setverbose"      : (self.HandleCmdVerbose,      True ),
            "setquiet"        : (self.HandleCmdQuiet,        True ),
            "setlogfile"      : (self.HandleCmdLogfile,      True ),
            "setlacppoll"     : (self.HandleCmdLacpPoll,     True ),
            "setpeertimeout"  : (self.HandleCmdPeerTimeout,  True ),
            "setsendtimeout"  : (self.HandleCmdSendTimeout,  True ),
            "setsendbufsize"  : (self.HandleCmdSendBufSize,  True ),
            "setpriority"     : (self.HandleCmdPriority,     True ),
            "setpeerlinkpoll" : (self.HandleCmdPeerLinkPoll, True ),
            "getdualifs"      : (self.HandleCmdDualIfs,      False),
            "getpeeralive"    : (self.HandleCmdPeerAlive,    False),
            "getclagid"       : (self.HandleCmdClagId,       False),
            "getclagrole"     : (self.HandleCmdClagRole,     False),
            "getpeerid"       : (self.HandleCmdPeerId,       False),
            "getpeerrole"     : (self.HandleCmdPeerRole,     False),
            "getbackuppeerid" : (self.HandleCmdBackupPeerId, False),
            "getbackuppeerrole" : (self.HandleCmdBackupPeerRole, False),
            "getpeerif"       : (self.HandleCmdPeerIf,       False),
            "getpeerip"       : (self.HandleCmdPeerIp,       False),
            "getbackupip"     : (self.HandleCmdGetBackupIp,  False),
            "getbackupactive" : (self.HandleCmdBackupActive, False),
            "getbackuprole"   : (self.HandleCmdBackupRole,   False),
            "getsysmac"       : (self.HandleCmdSysMac,       False),
            "getourlacpdb"    : (self.HandleCmdOurLacpDB,    False),
            "getpeerlacpdb"   : (self.HandleCmdPeerLacpDB,   False),
            "getpeerlacprate" : (self.HandleCmdPeerLacpRate, False),
            "getclagbonddb"   : (self.HandleCmdClagBondDB,   False),
            "getclagintfdb"   : (self.HandleCmdClagIntfDB,   False),
            "getourmacdb"     : (self.HandleCmdOurMacDB,     False),
            "getpeermacdb"    : (self.HandleCmdPeerMacDB,    False),
            "getourmcastdb"   : (self.HandleCmdOurMdbDB,     False),
            "getpeermcastdb"  : (self.HandleCmdPeerMdbDB,    False),
            "getourrportdb"   : (self.HandleCmdOurRportDB,   False),
            "getpeerrportdb"  : (self.HandleCmdPeerRportDB,  False),
            "getpeervlandb"   : (self.HandleCmdPeerVlanDB,   False),
            "getourvlandb"    : (self.HandleCmdOurVlanDB,    False),
            "getnllinkcache"  : (self.HandleCmdNlLinkCache,  False),
            "getintinfo"      : (self.HandleCmdIntInfo,      False),
            "getsummary"      : (self.HandleCmdSummary,      False),
            "getbackupinfo"   : (self.HandleCmdBackupInfo,   False),
            "getqinfo"        : (self.HandleCmdGetQInfo,     False),
            "logmsg"          : (self.HandleCmdLogMsg,       True ),
            "collectgarbage"  : (self.HandleCmdGarbage,      True ),
            "getall"          : (self.HandleCmdGetAll,       False),
            "dumpmstpstate"   : (self.HandleCmdDumpMstp,     False),
            "nogetall"        : (self.HandleCmdNoGetAll,     False),
            "startprofile"    : (self.HandleStartProfile,    False),
            "stopprofile"     : (self.HandleStopProfile,     False),
            "dumpprofile"     : (self.HandleDumpProfile,     False),
        }

        ucredStr = ourSock.getsockopt(socket.SOL_SOCKET, cmdsrv._SOCKET_PEERCRED, struct.calcsize("3I"))
        (peerPid, peerUid, peerGid) = struct.unpack("3I", ucredStr)
        canModify = peerUid == 0

        self.cmdOutputLock.acquire()
        if ourSock not in self.cmdOutputBufs:
            self.cmdOutputBufs[ourSock] = ""

        termChar = chr(0)
        cmdList = cmdStr.split()
        if cmdList[0].lower() in clagCmds:
            if clagCmds[cmdList[0].lower()][1] and not canModify:
                self.cmdOutputBufs[ourSock] += "ERROR You do not have sufficient privilege to execute this command: %s" % (cmdStr,)
            else:
                termChar = clagCmds[cmdList[0].lower()][0](cmdStr, ourSock, self.cmdOutputBufs)
        else:
            self.cmdOutputBufs[ourSock] += "ERROR Unknown command: %s" % (cmdStr,)

        self.cmdOutputBufs[ourSock] += termChar
        self.cmdOutputLock.release()

