#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2014 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import clag.clagthread
    import clag.datasync
    import copy
    import os
    import Queue
    import subprocess
    import threading
    import json
    import clag.util as util
    import xml.etree.ElementTree as ET
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

class fdbsync(clag.datasync.datasync):
    """
    This class handles everything required to synchronize the Bridge Forwarding
    Database, or FDB, between two CLAG peers.
    """

    # Work queue operation codes.
    _MAC_ADD        = 1
    _MAC_DELETE     = 2
    _MAC_REFRESH    = 3
    _MAC_MOVE       = 4

    _RECENT         = 10
                  
    def __init__(self, Daemon, Intf, Parser, Log, sendQueue):
        """
        Constructor: Initialize data structures, threads, etc.
        """
        super(fdbsync, self).__init__()

        #
        #   Save the parameters passed in
        #
        self.Daemon     = Daemon
        self.Intf       = Intf
        self.Parser     = Parser
        self.Log        = Log
        self.sendQueue  = sendQueue

        self.peerIf = self.Parser.args.peerIf
        if self.Intf.isSubIfName(self.peerIf):
            self.peerIf = self.Intf.GetMasterOfSubIf(self.peerIf)

        #
        #   Mac Database - The MAC addresses which are present on an interface  
        #   of a bridge, are tracked in this database, which is a dictionary.   
        #   The key is a tuple containing the MAC address and VLAN ID. On a VLAN
        #   unaware bridge, the VLAN ID will be the peer interface's VLAN ID.
        #   The value of each entry is a list containing six elements: the 
        #   local "base" interface name on which the address exists, the remote
        #   "base" interface name on which the address exists or None for
        #   locally learned addresses, the time the entry was last updated, the
        #   time the entry was added, the local IP for VxLAN or None for non-
        #   VxLAN interfaces and the VNI for VxLAN or 0xFFFFFFFF for non-VxLAN.
        #                                                                       
        #   {
        #                MAC          Vlan       localIf  remoteIf  updated added dst   vni
        #       ("00:00:00:00:00:02", 101 )  : [ "bond1", "bond11",   14,    30,  None, 0xFFFFFFFF ],
        #       ("00:00:00:00:00:03", 101 )  : [ "bond2", None,        0,    10,  None, 0xFFFFFFFF ],
        #       ("00:00:00:00:00:04", 102 )  : [ "bond3", "bond13",    0,    10,  None, 0xFFFFFFFF ],
        #       ("00:00:00:00:00:05", 102 )  : [ "bond4", None,        0,    12,  None, 0xFFFFFFFF ],
        #       ("00:00:00:00:00:06", 101 )  : [ "vx-75", None,      100,   500,  "36.0.0.1",   75 ]
        #   }
        #
        self.macDBLock  = threading.Lock()
        self.macDB      = {}
        self.syncDB     = {}

        #
        #   Hardware Work Queue - This queue contains MAC addresses which need  
        #   to be added or deleted to/from the kernel/hardware.  Each entry is a
        #   tuple containing an operation, the MAC address, a VLAN Id, the
        #   "base" interface, and the destination IP address (for vxlan).
        #
        #       ( _MAC_ADD, "00:00:00:00:00:02", 101, "bond1", None )
        #       ( _MAC_ADD, "00:00:00:00:00:02", 102, "vx-10", "36.0.0.1" )
        #
        self.hwWorkQueue = Queue.Queue()
        util.queueAddNameMap(self.hwWorkQueue, 'fdbHwWorkQueue')

        #
        #   Remote Delete Queue - This queue contains remote MAC addresses which
        #   have  been requested by the peer switch to be deleted. Before being
        #   deleted, we need to check if we have seen the address recently. The
        #   format of each queue element is a tuple containing the MAC address,
        #   VLAN Id, vni, "base" interface, and local IP for VxLAN.
        #
        #       ( "00:00:00:00:00:02", 101, 0xFFFFFFF, "bond1", None )
        #
        self.remoteDeleteQueue = Queue.Queue()
        util.queueAddNameMap(self.remoteDeleteQueue, 'fdbRemoteDeleteQueue')

        #
        #   FDB Work Queue - This queue contains lines from the "bridge fdb
        #   monitor" or "bridge fdb show" command output, which are awaiting
        #   processing.
        #
        self.fdbWorkQueue = Queue.Queue()
        util.queueAddNameMap(self.fdbWorkQueue, 'fdbWorkQueue')

        #
        #   The FDB Subprocess - The value returned by Popen for communicating
        #   with the "bridge fdb monitor" command.
        #
        self.fdbSubProcess = None

        #
        #   Start threads to handle hardwre updates, kernel notifications, etc.
        #
        self.keep_going = True
        self.stopEvent  = clag.clagthread.Event()

        self.refreshMacsThread = threading.Thread(None, self.RefreshMacsT, "RefreshMacs")
        self.refreshMacsThread.daemon = True
        self.refreshMacsThread.start()

        self.updateHwThread = threading.Thread(None, self.UpdateMacsInHwT, "UpdateMacsInHw")
        self.updateHwThread.daemon = True
        self.updateHwThread.start()

        self.remoteDeleteThread = threading.Thread(None, self.RemoteDeleteT, "RemoteDelete")
        self.remoteDeleteThread.daemon = True
        self.remoteDeleteThread.start()

        self.fdbMonitorThread = threading.Thread(None, self.FdbMonitorT, "FdbMonitor")
        self.fdbMonitorThread.daemon = True
        self.fdbMonitorThread.start()

        self.kernelMsgThread = threading.Thread(None, self.FdbKernelMsgT, "FdbKernelMsg")
        self.kernelMsgThread.daemon = True
        self.kernelMsgThread.start()

    def __del__(self):
        """
        Destructor:
        """
        self.SignalThreads()
        self.JoinThreads()

    def RefreshMacsT(self):
        try:
            self.RefreshMacs()
        except:
            self.Daemon.DumpTraceback()

    def UpdateMacsInHwT(self):
        try:
            self.UpdateMacsInHw()
        except:
            self.Daemon.DumpTraceback()

    def RemoteDeleteT(self):
        try:
            self.RemoteDelete()
        except:
            self.Daemon.DumpTraceback()

    def FdbMonitorT(self):
        try:
            self.FdbMonitor()
        except:
            self.Daemon.DumpTraceback()

    def FdbKernelMsgT(self):
        try:
            self.FdbKernelMsg()
        except:
            self.Daemon.DumpTraceback()

    def SignalThreads(self):
        self.keep_going = False
        self.stopEvent.set()
        if self.fdbSubProcess is not None:
            self.fdbSubProcess.terminate()
        try:
            self.hwWorkQueue.put_nowait((None, None, None, None, None))
            util.queueSetMaxLen('fdbHwWorkQueue')
        except Queue.Full:
            pass
        try:
            self.remoteDeleteQueue.put_nowait((None, None, None, None, None))
            util.queueSetMaxLen('fdbRemoteDeleteQueue')
        except Queue.Full:
            pass
        try:
            self.fdbWorkQueue.put_nowait(None)
            util.queueSetMaxLen('fdbWorkQueue')
        except Queue.Full:
            pass

    def JoinThreads(self):
        self.refreshMacsThread.join()
        self.updateHwThread.join()
        self.remoteDeleteThread.join()
        self.fdbMonitorThread.join()
        self.kernelMsgThread.join()

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_FDB_SYNC, logStr)

    def HwWorkQueuePut(self, element):
        if self.keep_going:
            self.hwWorkQueue.put(element)
            util.queueSetMaxLen('fdbHwWorkQueue')

    def RemoteDeleteQueuePut(self, element):
        if self.keep_going:
            self.remoteDeleteQueue.put(element)
            util.queueSetMaxLen('fdbRemoteDeleteQueue')

    def __str__(self):
        """
        Return local information in human-readable form
        """
        macDb = []
        self.macDBLock.acquire()
        for (mac, vlanId), [lif, rif, updated, added, dst, vni] in self.macDB.iteritems():
            if not rif:
                macInfo = {}
                macInfo['outIf'] = lif
                macInfo['vlanId'] = vlanId
                macInfo['mac'] = mac
                if dst:
                    macInfo['dst'] = dst
                if vni != 0xFFFFFFFF:
                    macInfo['vni'] = vni
                macDb.append(macInfo)
        self.macDBLock.release()
        return json.dumps(macDb, sort_keys=True, indent=4)

    def EncodedOurStr(self):
        return self.__str__()

    def EncodedPeerStr(self):
        """
        Return peer information in human-readable form
        """
        macDb = []
        self.macDBLock.acquire()
        for (mac, vlanId), [lif, rif, updated, added, dst, vni] in self.macDB.iteritems():
            if rif:
                macInfo = {}
                macInfo['outIf'] = lif
                macInfo['vlanId'] = vlanId
                macInfo['mac'] = mac
                if dst:
                    macInfo['dst'] = dst
                if vni != 0xFFFFFFFF:
                    macInfo['vni'] = vni
                macDb.append(macInfo)
        self.macDBLock.release()
        return json.dumps(macDb, sort_keys=True, indent=4)

    def CollectLocalInfo(self):
        """
        This function is called on a periodic interval to gather up any 
        information on the local switch which needs to be collected prior to
        sending that information to the peer switch.
        """
        pass

    def RefreshMacs(self):
        """
        This function executes periodically to refresh the MAC age of remote
        addresses.
        """
        self.log_debug("Beginning execution of the thread RefreshMacs")
        lastAgeRefresh = int(os.times()[4])
        while self.keep_going:

            shortestMaxAge = 300
            for br in self.Intf.GetAllBridges():
                maxAge = self.Intf.GetBridgeAgeingTime(br)
                if maxAge is not None and maxAge < shortestMaxAge:
                    shortestMaxAge = maxAge

            now = int(os.times()[4])
            if (now - lastAgeRefresh) > (shortestMaxAge * 0.45):
                self.log_debug("Adding entry to the fdbWorkQueue: %s" % ("DoRefreshMacs",))
                self.fdbWorkQueue.put("DoRefreshMacs")
                lastAgeRefresh = now

            self.log_debug("Waiting %d seconds before continuing execution of the thread RefreshMacs" % (self.Parser.getPeriodicRun(),))
            self.stopEvent.wait(self.Parser.getPeriodicRun())
            self.log_debug("Executing the thread RefreshMacs")

        self.log_debug("Finished execution of the thread RefreshMacs")

    def DoRefreshMacs(self, now, lastAgeRefresh):
        self.log_debug("Beginning RefreshMacs")
        dualIfs = self.Intf.GetDualIfs()
        k_macDB = self.GetKernelFdb()
        self.SyncCurrentState(k_macDB)
        self.macDBLock.acquire()
        c_macDB = copy.deepcopy(self.macDB)
        self.macDBLock.release()

        for (k_mac, k_vlanId), [k_if, k_age, k_dst, k_vni] in k_macDB.items():
            [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] = c_macDB.get((k_mac, k_vlanId), [None, None, None, None, None, None])
            if c_rif and (now - lastAgeRefresh) < (k_age + fdbsync._RECENT):
                if k_if not in dualIfs:
                    k_if = self.peerIf
                self.log_debug("Refreshing age of remote address %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, k_if, k_dst))
                self.macDBLock.acquire()
                self.macDB[(k_mac, k_vlanId)] = [k_if, c_rif, now, c_added, k_dst, k_vni]
                self.macDBLock.release()
                self.HwWorkQueuePut((fdbsync._MAC_REFRESH, k_mac, k_vlanId, k_if, k_dst))
            c_macDB.pop((k_mac, k_vlanId), None)

        # Any addresses left are missed kernel deletes, re-add them to the kernel
        for (c_mac, c_vlanId), [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] in c_macDB.items():
            if c_rif:
                self.log_debug("Sync re-adding removed remote address %s vlan %s vni %s dev %s dst %s" % (c_mac, c_vlanId, c_vni, c_lif, c_dst))
                self.macDBLock.acquire()
                self.macDB[(c_mac, c_vlanId)] = [c_lif, c_rif, now, c_added, c_dst, c_vni]
                self.macDBLock.release()
                self.HwWorkQueuePut((fdbsync._MAC_ADD, c_mac, c_vlanId, c_lif, c_dst))

        self.log_debug("Ending RefreshMacs")

    def SyncCurrentState(self, k_macDB):
        '''
        Given the current state of the FDB, make sure that the clag database
        is in sync with it. Also, make sure that the peer is in sync with these
        addresses.
        '''
        self.log_debug("Beginning sync of local addresses")

        self.macDBLock.acquire()
        c_macDB = copy.deepcopy(self.macDB)
        self.macDBLock.release()
        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        ET.SubElement(root, "fdb-start-sync")
        xmlEntries = 0

        now = int(os.times()[4])
        for (k_mac, k_vlanId), [k_if, k_age, k_dst, k_vni] in k_macDB.items():
            [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] = c_macDB.get((k_mac, k_vlanId), [None, None, now, now, None, None])
            # Is it a local address, not on the peer interface?
            if not c_rif and k_if != self.peerIf:
                # Update our clag database with this address
                self.log_debug("Syncing local address %s vlan %s vni %s dev %s dst %s, c_lif %s" % (k_mac, k_vlanId, k_vni, k_if, k_dst, c_lif))
                self.macDBLock.acquire()
                self.macDB[(k_mac, k_vlanId)] = [k_if, None, c_updated, c_added, k_dst, k_vni]
                self.macDBLock.release()
                c_macDB.pop((k_mac, k_vlanId), None)

                # Inform the peer of this address.  
                ET.SubElement(root, "fdb", {"op" : "sync", "mac" : k_mac, "vlanId" : str(k_vlanId), "vni" : str(k_vni), "if" : k_if, "dst" : str(k_dst)})
                xmlEntries += 1
                if xmlEntries >= 2000:
                    xmlStr = ET.tostring(root)
                    self.sendQueue.put(xmlStr)
                    root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                    xmlEntries = 0

        # Any addresses left are missed kernel deletes.
        self.macDBLock.acquire()
        if len(c_macDB):
            brPeerVlanMap = self.Intf.GetBridgePeerVlanMap()
        for (c_mac, c_vlanId), [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] in c_macDB.items():
            brIf = self.Intf.GetBridgeOfMember(c_lif)
            if not c_rif and (self.Intf.isVlanFilteringBridge(brIf) or brIf in brPeerVlanMap):
                self.log_debug("Sync removing local address %s vlan %s vni %s dev %s dst %s" % (c_mac, c_vlanId, c_vni, c_lif, c_dst))
                self.macDB.pop((c_mac, c_vlanId), None)
        self.macDBLock.release()

        ET.SubElement(root, "fdb-stop-sync")
        xmlStr = ET.tostring(root)
        self.sendQueue.put(xmlStr)

        self.log_debug("Ending sync of local addresses")

    def EncodeXmlInfo(self, parent):
        """
        Encodes the local information as XML for sending to the peer on a
        periodic basis.
        """
        pass

    def FdbMonitor(self):
        """
        This is an independent thread of the script. It is responsible for 
        providing the output of the "bridge fdb monitor" command to the
        kernel message thread through the fdbWorkQueue.
        """
        self.log_debug("Beginning execution of the thread FdbMonitor")

        while self.keep_going:
            try:
                self.log_debug("Opening fdbSubProcess")
                self.fdbSubProcess = subprocess.Popen(['/sbin/bridge', 'monitor', 'fdb'], 
                                                      stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            except OSError:
                self.Log.log_error("Unable to execute the '/sbin/bridge monitor fdb' command. Unable to sync MAC addresses.")
                self.fdbSubProcess = None
                self.Daemon.SignalAllThreadsToStop()

            # Get the current FDB, just to kick things off.
            k_macDB = self.GetKernelFdb()
            self.SyncCurrentState(k_macDB)

            # Now use the monitor command to see all FDB modifications
            for line in iter(self.fdbSubProcess.stdout.readline, b''):
                self.log_debug("Adding entry to the fdbWorkQueue: %s" % (line.strip(),))
                self.fdbWorkQueue.put(line.strip())
                util.queueSetMaxLen('fdbWorkQueue')

        self.log_debug("Finished execution of the thread FdbMonitor")

    def PrintElement(self, node, level):
        self.log_debug("%s%s: %s: %s" % (" " * (level*3), node.tag, node.text, str(node.attrib)))
        for child in node.findall('*'):
            self.PrintElement(child, level+1)

    def GetWorkQueueDelayed(self, delay=0.001):
        '''
        Returns the next entry on the FDB work queue, waiting up to the specified
        delay (default 1 millisecond) for an entry.
        '''
        queueEmpty = False
        try:
            macLineStr = self.fdbWorkQueue.get(True, delay)
        except Queue.Empty:
            macLineStr = ""
            queueEmpty = True
        return (macLineStr, queueEmpty)

    def FdbKernelMsg(self):
        """
        This is an independent thread of the script. It is responsible for
        processing entries on the fdbWorkQueue, which may result in modifications
        to the macDB and messages sent to the peer.
        """
        self.log_debug("Beginning execution of the thread FdbKernelMsg")                  

        lastAgeRefresh = int(os.times()[4])
        while self.keep_going:
            self.log_debug("Waiting for an event before continuing execution of the thread FdbKernelMsg")
            macLineStr = self.fdbWorkQueue.get()

            if self.keep_going:
                queueEmpty = False
                vlanMap = self.Intf.GetSubIfsVlanMap()
                brPeerVlanMap = self.Intf.GetBridgePeerVlanMap()
                pvidMap = self.Intf.GetPvidMap()
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0
                peerDown = self.Intf.GetOperStateFromSysFs(self.peerIf) != "up"
                while not queueEmpty and self.keep_going:
                    if macLineStr == "DoRefreshMacs":
                        now = int(os.times()[4])
                        self.DoRefreshMacs(now, lastAgeRefresh)
                        lastAgeRefresh = now
                    elif not macLineStr.endswith("permanent") and not macLineStr.endswith("static"):
                        self.log_debug("Processing entry: %s" % (macLineStr,))
                        k_add = True
                        if macLineStr.startswith("Deleted "):
                            macLineStr = macLineStr[8:]
                            k_add = False
                        macLineList = macLineStr.split()

                        # VLAN Unaware:
                        # 00:aa:00:00:00:01 dev vx-14567103 dst 36.0.0.13 vlan 65535 self 
                        # 00:aa:00:00:00:01 dev vx-14567103 vlan 0 master br-14567103
                        # 00:aa:00:00:00:01 dev t2h_c-4.1010 vlan 0 master br-14567103
                        #
                        # VLAN Aware:
                        # 00:aa:00:00:00:01 dev vx-14567103 dst 36.0.0.13 vlan 65535 self 
                        # 00:aa:00:00:00:01 dev vx-14567103 vlan 1003 master vx_aware-1
                        # 00:aa:00:00:00:02 dev t2h_c-4 vlan 1003 master vx_aware-1

                        if macLineList:
                            k_mac = macLineList[0]

                            try: # to get the interface MAC was learned upon
                                devIdx = macLineList.index('dev')
                                k_bif = macLineList[devIdx+1]
                            except (ValueError, IndexError): # No 'dev' keyword or interface following it
                                (macLineStr, queueEmpty) = self.GetWorkQueueDelayed()
                                continue
                            k_if = vlanMap.get(k_bif, (k_bif, None))[0]

                            vxlanInfo = self.Parser.GetVxLanInfo(k_if)
                            try: # to get the remote IP address (vxlan)
                                dstIdx = macLineList.index('dst')
                                k_dst = macLineList[dstIdx+1]
                                if not vxlanInfo or not k_add:
                                    (macLineStr, queueEmpty) = self.GetWorkQueueDelayed()
                                    continue
                                k_vni = vxlanInfo[0]
                            except (ValueError, IndexError):
                                if vxlanInfo: 
                                    if k_add: # Ignore VxLAN on bridge driver for adds
                                        (macLineStr, queueEmpty) = self.GetWorkQueueDelayed()
                                        continue
                                    else:
                                        k_dst = vxlanInfo[1]
                                        k_vni = vxlanInfo[0]
                                else:
                                    k_dst = None
                                    k_vni = 0xFFFFFFFF

                            k_vlanId = 0xFFFF
                            if not k_dst:
                                try: # to get the vlanid
                                    vlanIdx = macLineList.index('vlan')
                                    k_vlanId = int(macLineList[vlanIdx+1])
                                except (ValueError, IndexError):
                                    k_vlanId = 0

                            if not k_vlanId:
                                try: # to get the master interface
                                    mstrIdx = macLineList.index('master')
                                    k_master = macLineList[mstrIdx+1]
                                except (ValueError, IndexError):
                                    (macLineStr, queueEmpty) = self.GetWorkQueueDelayed()
                                    continue
                                k_vlanId = brPeerVlanMap.get(k_master, None)
                                if k_vlanId is None:
                                    (macLineStr, queueEmpty) = self.GetWorkQueueDelayed()
                                    continue
                            elif k_vlanId == 0xFFFF:
                                k_master = self.Intf.GetBridgeOfMember(k_if)
                                k_vlanId = brPeerVlanMap.get(k_master, pvidMap.get(k_if))
                                if k_vlanId is None:
                                    (macLineStr, queueEmpty) = self.GetWorkQueueDelayed()
                                    continue

                            self.macDBLock.acquire()
                            now = int(os.times()[4])
                            # Is this a new address we don't know anything about yet?
                            if (k_mac, k_vlanId) not in self.macDB:
                                if k_add and k_if != self.peerIf:
                                    self.log_debug("Adding new local address %s vlan %s vni %s dev %s" % (k_mac, k_vlanId, k_vni, k_if))
                                    self.macDB[(k_mac, k_vlanId)] = [k_if, None, now, now, k_dst, k_vni]
                                    ET.SubElement(root, "fdb", {"op" : "add", "mac" : k_mac, "vlanId" : str(k_vlanId), "vni" : str(k_vni), "if" : k_if, "dst" : str(k_dst)})
                                    xmlEntries += 1
                            else:
                                # We've seen this address before. Is it a local address?
                                [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] = self.macDB[(k_mac, k_vlanId)]
                                if not c_rif:
                                    if k_add:
                                        # Is the address moving to a different port?
                                        if k_if != c_lif or k_dst != c_dst:
                                            if k_if == self.peerIf: 
                                                if (now - c_added) > fdbsync._RECENT:
                                                    self.log_debug("Removing existing address (moved to peer) %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, c_lif, c_dst))
                                                    del self.macDB[(k_mac, k_vlanId)]
                                                else:
                                                    self.log_debug("Re-adding existing local address %s vlan %s vni %s dev %s" % (k_mac, k_vlanId, k_vni, c_lif))
                                                    self.macDB[(k_mac, k_vlanId)] = [c_lif, None, c_updated, c_added, c_dst, c_vni]
                                                    self.HwWorkQueuePut((fdbsync._MAC_ADD, k_mac, k_vlanId, c_lif, c_dst))
                                            else:
                                                self.log_debug("Moving existing local address %s vlan %s vni %s dst %s from %s to %s" % (k_mac, k_vlanId, k_vni, k_dst, c_lif, k_if))
                                                self.macDB[(k_mac, k_vlanId)] = [k_if, None, now, now, k_dst, k_vni]
# TODO: XXX Look at all ET.SubElements() calls for VLAN ID of VxLAN devices. May have to make them 0xFFFF before sending them across.
                                                ET.SubElement(root, "fdb", {"op" : "add", "mac" : k_mac, "vlanId" : str(k_vlanId), "vni" : str(k_vni), "if" : k_if, "dst" : str(k_dst)})
                                                xmlEntries += 1
                                    else:
                                        self.log_debug("Removing existing local address %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, c_lif, c_dst))
                                        del self.macDB[(k_mac, k_vlanId)]
                                        ET.SubElement(root, "fdb", {"op" : "delete", "mac" : k_mac, "vlanId" : str(k_vlanId), "vni" : str(k_vni), "if" : c_lif, "dst" : str(c_dst)})
                                        xmlEntries += 1
                                else:
                                    # Remote address.
                                    if k_add:
                                        if (now - c_added) > fdbsync._RECENT and k_if != self.peerIf:
                                            self.log_debug("Moving existing remote address to local %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, k_if, k_dst))
                                            self.macDB[(k_mac, k_vlanId)] = [k_if, None, c_updated, c_added, k_dst, k_vni]
                                            ET.SubElement(root, "fdb", {"op" : "add", "mac" : k_mac, "vlanId" : str(k_vlanId), "vni" : str(k_vni), "if" : k_if, "dst" : str(k_dst)})
                                            xmlEntries += 1
                                        else:
                                            if k_if != c_lif:
                                                self.log_debug("Re-adding existing remote address %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, c_lif, c_dst))
                                                self.macDB[(k_mac, k_vlanId)] = [c_lif, c_rif, c_updated, c_added, c_dst, c_vni]
                                                self.HwWorkQueuePut((fdbsync._MAC_ADD, k_mac, k_vlanId, c_lif, c_dst))
                                    else:
                                        if k_if == self.peerIf and peerDown:
                                            self.log_debug("Removing remote address peer-down %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, c_lif, c_dst))
                                            del self.macDB[(k_mac, k_vlanId)]
                                        else:
                                            if self.Intf.GetOperStateFromSysFs(k_bif) == "up":
                                                self.log_debug("Re-adding deleted remote address %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, c_lif, c_dst))
                                                self.macDB[(k_mac, k_vlanId)] = [c_lif, c_rif, now, c_added, c_dst, c_vni]
                                                self.HwWorkQueuePut((fdbsync._MAC_ADD, k_mac, k_vlanId, c_lif, c_dst))
                                            else:
                                                if k_vni == 0xFFFFFFFF:
                                                    self.log_debug("Remote address, interface down, move address %s vlan %s vni %s dev %s to the peer" % (k_mac, k_vlanId, c_vni, c_lif))
                                                    self.macDB[(k_mac, k_vlanId)] = [self.peerIf, c_rif, now, now, None, k_vni]
                                                    self.HwWorkQueuePut((fdbsync._MAC_MOVE, k_mac, k_vlanId, self.peerIf, None))
                                                else:
                                                    self.log_debug("Removing remote address intf not up %s vlan %s vni %s dev %s dst %s" % (k_mac, k_vlanId, k_vni, c_lif, c_dst))
                                                    del self.macDB[(k_mac, k_vlanId)]

                            self.macDBLock.release()

                    if xmlEntries >= 2000 and self.keep_going:
                        xmlStr = ET.tostring(root)
                        self.sendQueue.put(xmlStr)
                        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                        xmlEntries = 0

                    (macLineStr, queueEmpty) = self.GetWorkQueueDelayed()

                if xmlEntries and self.keep_going:
                    #self.PrintElement(root, 0)
                    xmlStr = ET.tostring(root)
                    self.sendQueue.put(xmlStr)

        self.log_debug("Finished execution of the thread FdbKernelMsg")

    def DecodeXmlInfo(self, root):
        """
        Decodes XML-encoded information sent from the peer.

        root - The root element of the XML tree
        """
        # Get the current list of dual connected hosts
        self.log_debug("Beginning DecodeXmlInfo")
        dualIfs    = self.Intf.GetDualIfs()
        revDualIfs = { v : k for k , v in dualIfs.items() }
        new_root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        xmlEntries = 0

        self.macDBLock.acquire()
        if len(root.findall('fdb-start-sync')):
            self.log_debug("Starting sync of remote addresses")
            self.syncDB = copy.deepcopy(self.macDB)

        now = int(os.times()[4])

# TODO: XXX Need to handle the case where an older peer sends a 0 or 0xFFFF vlanid (need to convert to valid value)
        for fdb in root.findall('fdb'):
            r_add = fdb.get('op') in [ "add", "sync" ]
            r_mac = fdb.get('mac')
            r_vlanId = fdb.get('vlanId')
            if r_vlanId == "None":
                r_vlanId = None
            if r_vlanId is not None:
                r_vlanId = int(r_vlanId)
            r_vni = fdb.get('vni')
            if r_vni == "None":
                r_vni = None
            if r_vni is not None:
                r_vni = int(r_vni)
            r_dst = fdb.get('dst')
            if r_dst == "None":
                r_dst = None
            r_peerIf = fdb.get('if')
            r_if = revDualIfs.get(r_peerIf, self.peerIf)
            if fdb.get('op') == "sync":
                self.syncDB.pop((r_mac, r_vlanId), None)

            if (r_mac, r_vlanId) not in self.macDB:
                if r_add:
                    self.log_debug("Adding new remote address %s vlan %s vni %s dev %s dst %s" % (r_mac, r_vlanId, r_vni, r_if, r_dst))
                    self.macDB[(r_mac, r_vlanId)] = [r_if, r_peerIf, now, now, r_dst, r_vni]
                    self.HwWorkQueuePut((fdbsync._MAC_ADD, r_mac, r_vlanId, r_if, r_dst))
            else:
                [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] = self.macDB[(r_mac, r_vlanId)]
                if not c_rif:
                    if r_add:
                        if (now - c_added) > fdbsync._RECENT or self.Parser.clagRole != self.Parser._ROLE_PRIMARY:
                            self.log_debug("Moving existing local address to remote %s vlan %s vni %s dev %s dst %s" % (r_mac, r_vlanId, r_vni, c_lif, r_dst))
                            self.macDB[(r_mac, r_vlanId)] = [r_if, r_peerIf, now, now, r_dst, r_vni]
                            if c_lif != r_if or c_dst != r_dst:
                                self.HwWorkQueuePut((fdbsync._MAC_MOVE, r_mac, r_vlanId, r_if, r_dst))
                    else:
                        self.log_debug("Telling peer to re-add address he deleted %s vlan %s vni %s dev %s dst %s" % (r_mac, r_vlanId, r_vni, c_lif, c_dst))
                        ET.SubElement(new_root, "fdb", {"op" : "add", "mac" : r_mac, "vlanId" : str(r_vlanId), "vni" : str(r_vni), "if" : c_lif, "dst" : str(c_dst)})
                        xmlEntries += 1
                else:
                    if r_add:
                        if c_lif != r_if or c_dst != r_dst:
                            self.log_debug("Moving remote address %s vlan %s vni %s dev %s dst %s to %s" % (r_mac, r_vlanId, r_vni, c_lif, r_dst, r_if))
                            self.macDB[(r_mac, r_vlanId)] = [r_if, r_peerIf, now, now, r_dst, r_vni]
                            self.HwWorkQueuePut((fdbsync._MAC_MOVE, r_mac, r_vlanId, r_if, r_dst))
                    else:
                        if r_if == self.peerIf and r_if == c_lif:
                            self.log_debug("Removing remote address %s vlan %s vni %s dev %s" % (r_mac, r_vlanId, r_if, r_vni))
                            del self.macDB[(r_mac, r_vlanId)]
                            self.HwWorkQueuePut((fdbsync._MAC_DELETE, r_mac, r_vlanId, r_if, r_dst))
                        else:
                            self.log_debug("Enqueueing remote delete %s vlan %s vni %s dev %s dst %s" % (r_mac, r_vlanId, r_vni, c_lif, r_dst))
                            self.RemoteDeleteQueuePut((r_mac, r_vlanId, r_vni, c_lif, r_dst))


        if len(root.findall('fdb-stop-sync')):
            self.log_debug("Stoping sync of remote addresses")
            for (s_mac, s_vlanId), [s_lif, s_rif, s_updated, s_added, s_dst, s_vni] in self.syncDB.items():
                if s_rif:
                    self.log_debug("Sync Removing remote address %s vlan %s vni %s dev %s" % (s_mac, s_vlanId, s_vni, s_lif))
                    self.macDB.pop((s_mac, s_vlanId), None)
                    self.HwWorkQueuePut((fdbsync._MAC_DELETE, s_mac, s_vlanId, s_lif, s_dst))

        self.macDBLock.release()

        if xmlEntries and self.keep_going:
            xmlStr = ET.tostring(new_root)
            self.sendQueue.put(xmlStr)
        self.log_debug("Ending DecodeXmlInfo")

    def GetKernelFdb(self):
        '''
        Retrieve the kernel's forwarding database. The database is a python
        dictionary whose key is a tuple of the mac address and vlan id. The 
        data of each entry is a list, with the first element the interface
        on which the mac/vlan exists and the second element is the number of
        seconds since the address was last seen.
        '''
        vlanMap = self.Intf.GetSubIfsVlanMap()
        brPeerVlanMap = self.Intf.GetBridgePeerVlanMap()
        pvidMap = self.Intf.GetPvidMap()
        k_macDB = {}

        try:
            macTableStr = subprocess.check_output(["/sbin/bridge", "-s", "fdb", "show"])
        except subprocess.CalledProcessError as e:
            self.Log.log_error("Error getting mac address table (bridge -s fdb show). %s" % (e,))
            macTableStr = ""

        for macLineStr in macTableStr.splitlines():
            if not macLineStr.endswith("permanent") and not macLineStr.endswith("static"):
                macLineList = macLineStr.split()

                if macLineList:
                    k_mac = macLineList[0]

                    try: # to get the interface MAC was learned upon
                        devIdx = macLineList.index('dev')
                        k_if = macLineList[devIdx+1]
                    except (ValueError, IndexError): # No 'dev' keyword or interface following it
                        continue
                    k_if = vlanMap.get(k_if, (k_if, None))[0]

                    vxlanInfo = self.Parser.GetVxLanInfo(k_if)
                    try: # to get the remote IP address (vxlan)
                        dstIdx = macLineList.index('dst')
                        k_dst = macLineList[dstIdx+1]
                        if not vxlanInfo:
                            continue
                        k_vni = vxlanInfo[0]
                    except (ValueError, IndexError):
                        if vxlanInfo:
                            continue
                        k_dst = None
                        k_vni = 0xFFFFFFFF

                    k_vlanId = 0xFFFF
                    if not k_dst:
                        try: # to get the vlanid
                            vlanIdx = macLineList.index('vlan')
                            k_vlanId = int(macLineList[vlanIdx+1])
                        except (ValueError, IndexError):
                            k_vlanId = 0

                    if not k_vlanId:
                        try: # to get the master interface
                            mstrIdx = macLineList.index('master')
                            k_master = macLineList[mstrIdx+1]
                        except (ValueError, IndexError):
                            continue
                        k_vlanId = brPeerVlanMap.get(k_master, None)
                        if k_vlanId is None:
                            continue
                    elif k_vlanId == 0xFFFF:
                        k_master = self.Intf.GetBridgeOfMember(k_if)
                        k_vlanId = brPeerVlanMap.get(k_master, pvidMap.get(k_if))
                        if k_vlanId is None:
                            continue

                    try: # to get the age
                        usedIdx = macLineList.index('used')
                        k_age = int(macLineList[usedIdx+1].split('/')[1])
                    except ValueError:
                        k_age = 0

                    k_macDB[(k_mac, k_vlanId)] = [k_if, k_age, k_dst, k_vni]

        return k_macDB

    def RemoteDelete(self):
        """
        This is an independent thread of the script. It is responsible for
        processing entries on the RemoteDeleteQueue, which handles requests from
        the peer to delete an address.
        """
        self.log_debug("Beginning execution of the thread RemoteDelete")                  

        while self.keep_going:
            self.log_debug("Waiting for an event before continuing execution of the thread RemoteDelete")
            (r_mac, r_vlanId, r_vni, r_if, r_dst) = self.remoteDeleteQueue.get()

            if self.keep_going:
                queueEmpty = False
                dualIfs = self.Intf.GetDualIfs()
                now = int(os.times()[4])
                k_macDB = self.GetKernelFdb()
                root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                xmlEntries = 0
                while not queueEmpty and self.keep_going:
                    self.macDBLock.acquire()
                    [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] = self.macDB.get((r_mac, r_vlanId), [None, None, None, None, None, None])
                    if c_rif:
                        [k_if, k_age, k_dst, k_vni] = k_macDB.get((r_mac, r_vlanId), [None, 0, None, 0xFFFFFFFF])
                        if k_if in dualIfs:
                            self.log_debug("Processing delete candidate: %s vlan %s vni %s dev %s. c_lif: %s, c_rif: %s, c_updated: %s, k_if: %s, k_age: %s" % (r_mac, r_vlanId, r_vni, r_if, c_lif, c_rif, c_updated, k_if, k_age))
                            if (k_age + fdbsync._RECENT) < (now - c_updated):
                                self.log_debug("Telling peer to re-add address (now local) %s vlan %s vni %s dev %s. k_age: %s, now: %s, c_updated: %s" % (r_mac, r_vlanId, r_vni, r_if, k_age, now, c_updated))
                                self.macDB[(r_mac, r_vlanId)] = [k_if, None, c_updated, c_added, k_dst, c_vni]
                                ET.SubElement(root, "fdb", {"op" : "add", "mac" : r_mac, "vlanId" : str(r_vlanId), "vni" : str(c_vni), "if" : k_if, "dst" : str(k_dst)})
                                xmlEntries += 1
                            else:
                                self.log_debug("Removing remote address %s vlan %s vni %s dev %s" % (r_mac, r_vlanId, r_vni, k_if))
                                del self.macDB[(r_mac, r_vlanId)]
                                self.HwWorkQueuePut((fdbsync._MAC_DELETE, r_mac, r_vlanId, k_if, k_dst))
                        else:
                            self.log_debug("Removing non-dual remote address %s vlan %s vni %s dev %s" % (r_mac, r_vlanId, r_vni, r_if))
                            del self.macDB[(r_mac, r_vlanId)]
                            self.HwWorkQueuePut((fdbsync._MAC_DELETE, r_mac, r_vlanId, r_if, r_dst))

                    self.macDBLock.release()

                    try:
                        (r_mac, r_vlanId, r_vni, r_if, r_dst) = self.remoteDeleteQueue.get(True, 0.001)
                    except Queue.Empty:
                        queueEmpty = True

                if xmlEntries and self.keep_going:
                    xmlStr = ET.tostring(root)
                    self.sendQueue.put(xmlStr)

        self.log_debug("Finished execution of the thread RemoteDelete")

    def DualIfChange(self, ifs, enable):
        """
        Called when an interface becomes or is no longer dual connected
        """
        if self.keep_going:
            root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
            xmlEntries = 0
            if enable:
                dualIfs    = self.Intf.GetDualIfs()
                revDualIfs = { v : k for k , v in dualIfs.items() }

                self.macDBLock.acquire()
                now = int(os.times()[4])
                for (c_mac, c_vlanId), [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] in self.macDB.items():
                    r_lif = revDualIfs.get(c_rif, None)
                    if r_lif in ifs:
                        self.log_debug("Dual connected, Moving address %s vlan %s vni %s dev %s to %s" % (c_mac, c_vlanId, c_vni, c_lif, r_lif))
                        self.macDB[(c_mac, c_vlanId)] = [r_lif, c_rif, now, now, c_dst, c_vni]
                        self.HwWorkQueuePut((fdbsync._MAC_MOVE, c_mac, c_vlanId, r_lif, c_dst))
                    if c_lif in ifs:
                        self.log_debug("Telling peer to re-add address (now dual connected) %s vlan %s vni %s dev %s." % (c_mac, c_vlanId, c_vni, c_lif))
                        ET.SubElement(root, "fdb", {"op" : "add", "mac" : c_mac, "vlanId" : str(c_vlanId), "vni" : str(c_vni), "if" : c_lif, "dst" : str(c_dst)})
                        xmlEntries += 1
                        if xmlEntries >= 2000 and self.keep_going:
                            xmlStr = ET.tostring(root)
                            self.sendQueue.put(xmlStr)
                            root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                            xmlEntries = 0
                self.macDBLock.release()

            else:
                # Interfaces are becoming singly connected
                self.macDBLock.acquire()
                operState = {}
                for s_if in ifs:
                    operState[s_if] = self.Intf.GetOperStateFromSysFs(s_if)
                now = int(os.times()[4])
                for (c_mac, c_vlanId), [c_lif, c_rif, c_updated, c_added, c_dst, c_vni] in self.macDB.items():
                    if c_lif in ifs:
                        if not self.Intf.isPeerAlive():
                            # If peer not alive and we are primary, make all addresses local.
                            if self.Parser.clagRole == self.Parser._ROLE_PRIMARY:
                                self.log_debug("Not dual connected, primary, peer down, make local %s vlan %s vni %s dev %s" % (c_mac, c_vlanId, c_vni, c_lif))
                                self.macDB[(c_mac, c_vlanId)] = [c_lif, None, c_updated, c_added, c_dst, c_vni]
                            # If peer not alive and we are secondary, flush all addresses.
                            else:
                                self.log_debug("Not dual connected, secondary, peer down, delete address %s vlan %s vni %s dev %s" % (c_mac, c_vlanId, c_vni, c_lif))
                                del self.macDB[(c_mac, c_vlanId)]
                                self.HwWorkQueuePut((fdbsync._MAC_DELETE, c_mac, c_vlanId, c_lif, c_dst))
                        else:
                            # If peer is alive and bond is down, move all remote MACs to peer. Remove all local MACs.
                            if operState[c_lif] != "up":
                                if c_rif is not None and c_vni != 0xFFFFFFFF:
                                        self.log_debug("Not dual connected, interface down, move address %s vlan %s vni %s dev %s to the peer" % (c_mac, c_vlanId, c_vni, c_lif))
                                        self.macDB[(c_mac, c_vlanId)] = [self.peerIf, c_rif, now, now, None, c_vni]
                                        self.HwWorkQueuePut((fdbsync._MAC_MOVE, c_mac, c_vlanId, self.peerIf, None))
                                else:
                                    self.log_debug("Not dual connected, interface down, delete local address %s vlan %s vni %s dev %s" % (c_mac, c_vlanId, c_vni, c_lif))
                                    del self.macDB[(c_mac, c_vlanId)]
                                    self.HwWorkQueuePut((fdbsync._MAC_DELETE, c_mac, c_vlanId, c_lif, None))
                                    if c_vni == 0xFFFFFFFF:
                                        ET.SubElement(root, "fdb", {"op" : "delete", "mac" : c_mac, "vlanId" : str(c_vlanId), "vni" : str(c_vni), "if" : c_lif, "dst" : "None"})
                                        xmlEntries += 1
                                        if xmlEntries >= 2000 and self.keep_going:
                                            xmlStr = ET.tostring(root)
                                            self.sendQueue.put(xmlStr)
                                            root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
                                            xmlEntries = 0
                self.macDBLock.release()

            if xmlEntries and self.keep_going:
                xmlStr = ET.tostring(root)
                self.sendQueue.put(xmlStr)

    def ClearPeerInfo(self):
        """
        Removes MAC addresses which are owned by the peer.
        """
        pass
#        if self.macDBLock.acquire(False):
#            for (mac, vlanId), [lif, rif, updated] in self.macDB.items():
#                if rif:
#                    self.HwWorkQueuePut((fdbsync._MAC_DELETE, mac, vlanId, lif, None))
#                    del self.macDB[(mac, vlanId)]
#            self.macDBLock.release()

    def ClearLocalInfo(self):
        """
        Removes MAC addresses which are owned by this switch.
        """
        self.macDBLock.acquire()
        for (mac, vlanId), [lif, rif, updated, added, dst, vni] in self.macDB.items():
            if not rif:
                self.HwWorkQueuePut((fdbsync._MAC_DELETE, mac, vlanId, lif, dst))
                del self.macDB[(mac, vlanId)]
        self.macDBLock.release()

    def UpdateMacsInHw(self):
        '''
        This function is an independent thread of the script. It compares the MAC
        address data base received from the peer (peerMacDB) with the mac addresses
        we have already programmed into the hardware. Any new addresses are added and
        any addresses no longer present on the peer are removed.
        '''
        fdb_cmd = {
            fdbsync._MAC_ADD     : ( "replace", "added to"     ),
            fdbsync._MAC_REFRESH : ( "replace", "refreshed on" ),
            fdbsync._MAC_MOVE    : ( "replace", "moved to"     ),
            fdbsync._MAC_DELETE  : ( "del",     "removed from" )
        }

        self.log_debug("Beginning execution of the thread UpdateMacsInHw")

        while self.keep_going:
            # Wait for a change in mac address data base received from the peer
            self.log_debug("Waiting for an event before continuing execution of the thread UpdateMacsInHw")
            (op, mac, vlanId, lif, dst) = self.hwWorkQueue.get()

            if self.keep_going:
                masterVlanMap = self.Intf.GetMasterIfVlanMap()
                peerVlanMap = self.Intf.GetPeerBridgeVlanMap()
                queueEmpty = False
                params = []
                vlanFilteringIf = {}
                while not queueEmpty and self.keep_going:
                    (fdb_op, fdb_str) = fdb_cmd.get(op, (None, None))
                    if fdb_op is not None:
                        pvlanId = peerVlanMap.get(vlanId, {}).get(lif)
                        intf = masterVlanMap.get(lif,{}).get(pvlanId,lif)
                        if intf is not None:
                            try:
                                vlanFiltering = vlanFilteringIf[intf]
                            except KeyError:
                                br = self.Intf.GetBridgeOfMember(intf)
                                if br is None:
                                    try:
                                        (op, mac, vlanId, lif, dst) = self.hwWorkQueue.get(True, 0.001)
                                    except Queue.Empty:
                                        queueEmpty = True
                                    continue
                                vlanFiltering = self.Intf.isVlanFilteringBridge(br)
                                vlanFilteringIf[intf] = vlanFiltering
                            vlanStr = " vlan %d" % (vlanId,) if vlanId and not pvlanId and vlanFiltering else ""
                            vxLanStr = " dst %s self" % (dst,) if dst else ""
                            useStr = " use" if op == fdbsync._MAC_REFRESH else ""
                            params += [(fdb_op, mac, intf, vlanStr, vxLanStr, useStr)]
                            self.Log.log_verbose("MAC address %s has been %s %s %s%s%s." % (mac, fdb_str, intf, vlanStr, vxLanStr, useStr))
                    else:
                        self.Log.log_error("Unknown op code in work queue: %s" % (op,))

                    try:
                        (op, mac, vlanId, lif, dst) = self.hwWorkQueue.get(True, 0.001)
                    except Queue.Empty:
                        queueEmpty = True

                self.Intf.ExecuteBatch(["/sbin/bridge", "-force", "-batch"],"fdb %s %s dev %s dynamic master%s%s%s", params)

        self.log_debug("Finished execution of the thread UpdateMacsInHw")

