#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2015 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import clag.clagthread
    import socket
    import threading
    import select
    import os
    import xml.etree.ElementTree as ET
    from ctypes import *
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

#-------------------------------------------------------------------------------
# This module checks the health of the peer switch via a backup IP address.
# 1. This IP address is typically the mgmt. IP address (but could also be an 
# inband/frontpanel IP). The backup IP config is optional (on account of upgrade
# from previous release but is recommended to avoid traffic loss on peer switch 
# death without "goodbye").
# 2. UDP is used for these health checks as UDP is connectionless (i.e. light 
# weight) and typically not blocked by mgmt network firewalls (on account of 
# traceroute UDP probes).
# 3. The healthcheck request is a light weight periodic hello request which is 
# turned around as a reply by clagd on the peer switch.
# 4. The backup state is only used if the backup was active at the time of peer 
# link down on the secondary switch. This takes care of the upgrade case where i
# the peer is not responding to the "new" hello requests. And the case were the 
# peer's mgmt IP was changed independently without updating the backup IP 
# config.
# 5. To avoid overwhelming the mgmt network the frequency of the 
# steady-state/periodic hellos has been kept low, one every 4 seconds. On a 
# potential peer death event (peer link down/peer-connection timeout) 
# healthCheck is moved to a rapid-detection mode (and the peer-down event is 
# deferred). In this mode 6 requests are sent 100ms apart. Even if one of the
# requests is responded too clagd assumes that the peer switch is alive (and 
# does not take over as primary; unless a goodbye is rxed). If no responses are
# rxed during rapid-detection the peer is considered dead and clagd takes over 
# the primary role.
#-------------------------------------------------------------------------------

libc_handle = cdll.LoadLibrary('libc.so.6')

class HealthCheck():
    def __init__(self, Daemon, Parser, Log, Intf):
        self.Daemon = Daemon
        self.Parser = Parser
        self.Log = Log
        self.Intf = Intf

        # params that cannot be changed on the fly
        self.peerHealthPort = self.Parser.args.backupPort

        # clag hello id - changes everytime clag is restarted
        self.helloId = os.getpid()
        # seq increments for every txed hello
        self.helloSeq = 0
        self.helloOutMaxReqs = 6
        # we can make the intervals configurable if needed; for now 4 seconds
        # and will take up to 24 seconds to declare peer dead. The time taken 
        # to declare backup inactive should be higher than the time it takes to 
        # detect peer link going down otherwise we will end up ignoring the 
        # backup state change on peer switch death.
        self.helloRegularInterval = 4
        self.helloRapidInterval = 0.1

        # Data shared by all the threads
        self.helloOutReqs = 0
        self.helloOutReqsLock = threading.Lock()
        self.backupActive = False

        # Election is run over the backup independently
        self.backupRoleAvailable = False
        self.backupRole = self.Parser.default_role
        self.peerId = [self.Parser.default_priority, "00:00:00:00:00:00"]
        self.peerRole = None

        # peerlink down triggers rapid detection (provided health is good at the
        # time of the event)
        self.helloRapidDetection = False
        self.keep_going = True

        # Reload timer (5min) - moves the clag bond out of error disabled state 
        # on clagd startup (if nothing is heard from the peer)
        # XXX - any way to avoid this thread? 
        self.helloReloadEvent = clag.clagthread.Event()
        self.helloReloadThread = threading.Thread(None, self.HelloReloadT, "HelloReloadT")
        self.helloReloadThread.daemon = True
        self.helloReloadThread.start()

        # helloTxT - tx thread - 
        # 1. Sends hello/echo-requests periodically
        # 2. Declares peerSwitch dead if outstanding (not replied to)
        #    requests reach max (helloOutMaxReqs)
        self.helloTxEvent = clag.clagthread.Event()
        try:
            self.helloTxSock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        except socket.error:
            self.Log.log_error("HealthCheck: Failed to create a udp socket")
            self.Daemon.SignalAllThreadsToStop()
            return

        # Try to bind the socket to the VRF if configured.
        self.backupVrfBindSuccess = False
        self.AttemptVrfBind()

        self.helloTT = threading.Thread(None, self.HelloTxT, "HelloTxT")
        self.helloTT.daemon = True
        self.helloTT.start()

        # helloRxT - rx thread - 
        # 1. Listens on peerHealthPort
        # 2. Replies to hello/echo-requests (simple turn around)
        # 3. Declares peerSwitch alive when a valid reply is rxed.
        #
        # Note: Using separate sockets for rx and tx (they are by seperate 
        # threads)
        # create a wake-up eventfd that can be used by the main thread to
        # wakeup the helloRx thread in case it needs to exit. Otherwise
        # the helloRx thread is blocked listening on the peerHealthPort
        self.helloRxWkupFd = libc_handle.eventfd(0, 0)
        try:
            self.helloRxSock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        except socket.error:
            self.Log.log_error("HealthCheck: Failed to create a udp socket")
            self.Daemon.SignalAllThreadsToStop()
            return

        # Bind the rx socket also to the VRF for now
        self.backupRxVrfBindSuccess = False
        self.AttemptRxVrfBind()

        try:
            self.helloRxSock.bind(('', self.peerHealthPort))
            self.helloRT = threading.Thread(None, self.HelloRxT, "HelloRxT")
            self.helloRT.daemon = True
            self.helloRT.start()
        except socket.error:
            self.Log.log_error("HealthCheck: Failed to bind to the backup udp port %d" % (self.peerHealthPort))
            self.Daemon.SignalAllThreadsToStop()
            return

    def __del__(self):
        # Signal the update thread to wake up and exit
        self.SignalThreads()
        self.JoinThreads()

    def HelloRxT(self):
        try:
            self.HelloRx()
            self.helloRxSock.close()
        except:
            self.Daemon.DumpTraceback()

    def HelloTxT(self):
        try:
            self.HelloTx()
        except:
            self.Daemon.DumpTraceback()

    def HelloReloadT(self):
        try:
            self.HelloReload()
        except:
            self.Daemon.DumpTraceback()

    def SignalThreads(self):
        self.keep_going = False
        self.helloTxEvent.set()
        try:
            self.helloReloadEvent.set()
        except AttributeError:
            pass
        try:
            os.write(self.helloRxWkupFd, c_ulonglong(1))
        except:
            self.Log.log_error("helloRx: failed to wakeup thread")

    def JoinThreads(self):
        try:
            self.helloRT.join()
            self.helloTT.join()
            self.helloReloadThread.join()
        except AttributeError:
            pass

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_HEALTH_CHECK, logStr)

    def PrintElement(self, node, level):
        self.log_debug("%s%s: %s: %s" % (" " * (level*3), node.tag, node.text, str(node.attrib)))
        for child in node.findall('*'):
            self.PrintElement(child, level+1)

    def AttemptVrfBind(self):
        if self.backupVrfBindSuccess or not self.Parser.args.backupVrf:
            return True

        self.log_debug("HealthCheck: bind socket to vrf %s" % self.Parser.args.backupVrf)
        try:
            self.helloTxSock.setsockopt(socket.SOL_SOCKET, 25, self.Parser.args.backupVrf)
        except socket.error:
            self.backupVrfBindSuccess = False
            self.Log.log_error("HealthCheck: Failed to bind socket to vrf %s" % self.Parser.args.backupVrf)
            return False

        self.backupVrfBindSuccess = True
        return True

    def AttemptRxVrfBind(self):
        if self.backupRxVrfBindSuccess or not self.Parser.args.backupVrf:
            return True

        self.log_debug("HealthCheck: bind rx socket to vrf %s" % self.Parser.args.backupVrf)
        try:
            self.helloRxSock.setsockopt(socket.SOL_SOCKET, 25, self.Parser.args.backupVrf)
        except socket.error:
            self.backupRxVrfBindSuccess = False
            self.Log.log_error("HealthCheck: Failed to bind rx socket to vrf %s" % self.Parser.args.backupVrf)
            return False

        self.backupRxVrfBindSuccess = True
        return True

    def HelloClearOutReqs(self):
        self.helloOutReqsLock.acquire()
        self.helloOutReqs = 0
        self.helloOutReqsLock.release()

    def HelloIncOutReqs(self):
        self.helloOutReqsLock.acquire()
        self.helloOutReqs = self.helloOutReqs + 1
        self.helloOutReqsLock.release()

    def HelloGetOutReqs(self):
        self.helloOutReqsLock.acquire()
        outReqs = self.helloOutReqs
        self.helloOutReqsLock.release()
        return outReqs

    def HelloBuildXmlMessage(self, echoType, echoId, seq):
        '''
        <clag_hello_data version="0.0.0">
            <echo type="request" id=helloId seq=1/>
            <echo type="request" id=helloId seq=1/>
            <i8023ad system_id="00:00:00:01:01:01"/>
            <clagid uniqueId="44:38:39:00:11:16" priority=32768 role=Primary/>
        </clag_hello_data>
        '''
        clagDataAttr = {'version' : self.Parser._ClagHelloVersion}
        root = ET.Element("clag_hello_data", clagDataAttr)
        ET.SubElement(root, "echo", {"type" : echoType, "id" : str(echoId), "seq" : str(seq)})
        ET.SubElement(root, "i8023ad", {"system_id" : self.Parser.args.sysMac})
        ET.SubElement(root, "clagid", {"priority" : str(self.Parser.clagId[0]), "uniqueId" : self.Parser.clagId[1], "role" : self.Parser.clagRole})
        #self.PrintElement(root, 0)
        return(ET.tostring(root))
        
    def HelloParseXmlMessage(self, xmlStr, addr):
        nakLogs = []
        # Convert the XML string into an ElementTree element
        try:
            root = ET.fromstring(xmlStr)
        except ET.ParseError:
            self.Log.log_error("HealthCheck: data from the peer is invalid:\n%s" % (xmlStr,))
            return

        # Is the root element named properly?
        if root is None or root.tag != "clag_hello_data":
            self.Log.log_error("HealthCheck: data from the peer has an invalid root tag:\n%s" % (xmlStr,))
            return

        # Is there a version attribute with the root element?
        verStr = root.get("version")
        if verStr is None:
            self.Log.log_error("HealthCheck: data from the peer does not include a version:\n%s" % (xmlStr,))
            return

        # Is the version something we understand?
        (peerMajVer, peerMinVer, peerSubVer) = verStr.split(".")
        (myMajVer,   myMinVer,   mySubVer)   = self.Parser._ClagHelloVersion.split(".")
        if int(peerMajVer) > int(myMajVer):
            self.Log.log_error("HealthCheck: version mismatch; My version: %s, Peer version: %s" % (self.Parser._ClagHelloVersion, verStr))
            return

        # if sysmac is present it needs to match with the local sysmac
        i8023ad = root.find('i8023ad')
        if i8023ad is not None:
            sysId = i8023ad.get('system_id')
            if sysId is None or sysId != self.Parser.args.sysMac:
                nakLogs.append("HealthCheck: system ID mismatch. Peer = %s, Our = %s" % ("None" if not sysId else sysId, self.Parser.args.sysMac))

        # Just parse clagId. We will only use it for election if it
        # was in a valid reply message.
        peerId = []
        peerRole = None
        clagIdNode = root.find('clagid')
        if clagIdNode is not None:
            peerId.append(int(clagIdNode.get('priority')))
            peerId.append(clagIdNode.get('uniqueId'))
            peerRole = clagIdNode.get('role')

        echoNode = root.find('echo')
        if echoNode is not None:
            echoType = echoNode.get('type')
            echoId = int(echoNode.get('id'), 0)
            seq = int(echoNode.get('seq'), 0)
            self.HelloProcEchoFromPeer(echoType, echoId, seq, addr, peerId, peerRole, nakLogs)

    def SetBackupRole(self, role):
        if self.backupRoleAvailable and self.backupRole == role:
            return
        self.Intf.protoDownLock.acquire()
        self.backupRole = role
        self.Log.log('HealthCheck: role via backup is %s' % \
                ('primary' if self.backupRole==self.Parser._ROLE_PRIMARY else 'secondary'))

        if not self.backupRoleAvailable:
            self.backupRoleAvailable = True

        # backupRole may be set before backupActive; in that case
        # SetBackupActive will update the proto_down state on all the 
        # interfaces
        if self.backupActive:
            params=[]
            for intf in self.Parser.GetClagIntfDB().values():
                params += intf.UpdateIntfProtoDown(batch=True)
            self.Intf.ExecuteBatch(["/sbin/ip", "-force", "-batch"], "link set dev %s protodown %s", params)
        self.Intf.protoDownLock.release()

    def IsBackupRoleAvailable(self):
        return self.backupActive and self.backupRoleAvailable

    def GetBackupRole(self):
        return self.backupRole

    def GetBackupActive(self):
        return self.backupActive

    def SetBackupActive(self, active):
        self.Intf.protoDownLock.acquire()
        self.backupActive = active
        if not self.backupActive:
            self.backupRoleAvailable = False
        self.Intf.SetProtoPeerOrBackupState()
        # backup role availability may have changed so re-evaluate
        params=[]
        for intf in self.Parser.GetClagIntfDB().values():
            params += intf.UpdateIntfProtoDown(batch=True)
        self.Intf.ExecuteBatch(["/sbin/ip", "-force", "-batch"], "link set dev %s protodown %s", params)
        self.Intf.protoDownLock.release()

    def HelloBackupReset(self):
        self.HelloClearOutReqs()
        if self.backupActive:
            self.Log.log('HealthCheck: backup ip changed')
            self.SetBackupActive(False)
        self.HelloRapidDetectionDone()

    def PeerRoleToVal(self, role):
        return 0 if role == self.Parser._ROLE_PRIMARY else 1

    def RunBackupElection(self):
        peerRoleVal = self.PeerRoleToVal(self.peerRole)
        roleVal = self.PeerRoleToVal(self.Parser.clagRole)
        # Same criteria for election as the one over the ISL. Only
        # difference is that we also include the peerRole determined over
        # ISL as a deciding criterion.
        if roleVal < peerRoleVal or \
            (roleVal == peerRoleVal and \
                (self.Parser.clagId[0] < self.peerId[0] or \
                (self.Parser.clagId[0] == self.peerId[0] and self.Parser.clagId[1] < self.peerId[1]))):
            self.SetBackupRole(self.Parser._ROLE_PRIMARY)
        else:
            self.SetBackupRole(self.Parser._ROLE_SECONDARY)

    def HelloProcEchoRep(self, peerId, peerRole):
        # On rxing an acceptable response we clear all pending requests
        self.HelloClearOutReqs()

        # perform election if the reply has the relevant info
        if peerId:
            self.peerId = peerId
            self.peerRole = peerRole
            self.RunBackupElection()

        if not self.backupActive:
            self.SetBackupActive(True)
            self.Log.log('HealthCheck: backup active')

        self.HelloRapidDetectionDone()
        
    def SwitchdIsNotReady(self):
        '''
        Check if switchd is ready. Use the status provided by systemd. This is
        really a check to see if the CPU can send and receive packets through
        the SWP ports.
        '''
        # Are we using knet interfaces? If so we are not ready when switchd is
        # activating (started but not ready)
        bcmKnetDir = "/sys/module/user_bde"
        if os.path.isdir(bcmKnetDir):
            return self.Parser.switchdState == self.Parser._SWITCHD_STARTING
        # Whithout knet, we are not ready in any other state than active.
        return self.Parser.switchdState != self.Parser._SWITCHD_READY

    def HelloProcEchoFromPeer(self, echoType, echoId, seq, addr, peerId, peerRole, nakLogs):
        if echoType == 'request':
            if self.SwitchdIsNotReady() and not self.Parser.args.vm:
                self.log_debug("RX message: ignored because switchd is not ready.")
                self.Intf.SetClagRole(self.Parser._ROLE_SECONDARY, "switchd not ready", True)
                return
            # reply to requests - simple turnaround
            xmlStr = self.HelloBuildXmlMessage('reply', echoId, seq)
            if self.keep_going:
                # if previous vrf binds had failed we try again; if it fails we will
                # supress transmission
                if not self.AttemptRxVrfBind():
                    return

                try:
                    self.helloRxSock.sendto(xmlStr, (addr[0], self.peerHealthPort))
                    self.log_debug("TX message: %s to %r" % (xmlStr, (addr[0], self.peerHealthPort)))
                except:
                    self.Log.log_error("helloRxSock send reply failed")
        elif echoType == 'reply':
            if nakLogs:
                #Some thing is wrong with the reply
                for nakLog in nakLogs:
                    self.Log.log_error(nakLog)
                self.Daemon.SignalAllThreadsToStop()
                return
            if echoId == self.helloId and seq == self.helloSeq:
                self.HelloProcEchoRep(peerId, peerRole)
        else:
            self.Log.log_error("HealthCheck: Rxed %s message from peer" % echoType)

    def HelloRx(self):
        # just receive all packets sent to the "helloPort"
        inputs = [self.helloRxSock, self.helloRxWkupFd]
        while self.keep_going and inputs:
            (readable, writeable, exceptional) = select.select(inputs, [], inputs)
            for errSock in exceptional:
                # TODO - what kind of UDP exceptions can we get
                self.Log.log_error('HealthCheck: %r exception ignored' % errSock)

            if self.keep_going:
                for readSock in readable:
                    data, addr = readSock.recvfrom(1024)
                    self.log_debug("RX message: %s from %r" % (data, addr))
                    self.HelloParseXmlMessage(data, addr)

    def HelloTx(self):
        while self.keep_going:
            self.log_debug("Executing thread HelloTx")
            # Check if the peer health has timed out
            outReqs = self.HelloGetOutReqs()
            if outReqs >= self.helloOutMaxReqs:
                if self.backupActive:
                    logstr = 'rapid' if self.helloRapidDetection else 'regular'
                    self.Log.log('HealthCheck: backup connection lost; %s' % logstr)

                    self.SetBackupActive(False)
                self.HelloRapidDetectionDone()

            # Send echo request
            healthIp = self.Parser.args.backupIp
            # if previous vrf binds had failed we try again; if it fails we will
            # supress transmission
            if healthIp and self.AttemptVrfBind():
                self.helloSeq = self.helloSeq + 1
                xmlStr = self.HelloBuildXmlMessage('request', self.helloId, self.helloSeq)
                try:
                    self.helloTxSock.sendto(xmlStr, (healthIp, self.peerHealthPort))
                except:
                    self.log_debug("helloTxSock send reply failed")
                self.HelloIncOutReqs()
                self.log_debug("TX message: %s to %r" % (xmlStr, (healthIp, self.peerHealthPort)))

            if self.helloRapidDetection:
                helloInterval = self.helloRapidInterval
            else:
                helloInterval = self.helloRegularInterval
            self.helloTxEvent.wait(helloInterval)
            self.helloTxEvent.clear()

    def HelloRapidDetectionSetup(self):
        if not self.helloRapidDetection:
            self.Log.log('HealthCheck: Rapid detection start')
            self.HelloClearOutReqs()
            self.helloRapidDetection = True
            self.helloTxEvent.set()

    def HelloRapidDetectionDone(self):
        if self.helloRapidDetection:
            logstr = 'active' if self.backupActive else 'inactive'
            self.Log.log('HealthCheck: Rapid detection done; backup %s\n' % logstr)
            self.helloRapidDetection = False
            self.Intf.ClearPeerDeathPending()

    def isPeerLinkDown(self):
        peerIf = self.Parser.args.peerIf
        if self.Intf.isSubIfName(peerIf):
            peerIf = self.Intf.GetMasterOfSubIf(peerIf)
        return self.Intf.GetOperStateFromSysFs(peerIf) != "up"

    def reloadExpired(self):
        return self.helloReloadEvent.is_set()

    def HelloReload(self):
        while self.keep_going:
            self.helloReloadEvent.wait(self.Parser.args.reloadTimer)
            self.helloReloadEvent.set()
            if self.keep_going:
                firstTime = True
                peerLinkIsDown = self.isPeerLinkDown()
                params = []
                self.Intf.protoDownLock.acquire()
                for intf in self.Parser.GetClagIntfDB().values():
                    if intf.protoDownFlags & intf._PROTO_DOWNF_INIT:
                        clearFlags = intf._PROTO_DOWNF_INIT
                        primaryRole = intf.GetProtoDownRole() == self.Parser._ROLE_PRIMARY
                        if intf.protoDownFlags & intf._PROTO_DOWNF_SINGLE_VXLAN and \
                           (primaryRole or peerLinkIsDown):
                            clearFlags |= intf._PROTO_DOWNF_SINGLE_VXLAN
                        if firstTime:
                            self.Log.log("HealthCheck: reload timeout.")
                            firstTime = False
                        params += intf.UpdateProtoDownFlags(0, clearFlags, batch=True)
                self.Intf.ExecuteBatch(["/sbin/ip", "-force", "-batch"], "link set dev %s protodown %s", params)
                self.Intf.protoDownLock.release()
                return
