#!/usr/bin/env python
#
# Copyright 2015 Cumulus Networks, Inc. All rights reserved.
# Author: Scott Feldman, sfeldma@cumulusnetworks.com
#
# Note: Currently ony link cache is supported.
#

from socket import AF_UNSPEC
from rtnetlink import *
import threading
import copy

class KCacheIface(Ifinfomsg):

    def __init__(self, ifi, flags, ifname, operstate, linkinfo, protinfo):
        memmove(addressof(self), addressof(ifi), sizeof(ifi))
        self.flags = flags
        self.ifname = ifname
        self.operstate = operstate
        self.linkinfo = linkinfo
        self.protinfo = protinfo
        self.stale = False

    def __eq__(self, other):
        # XXX - need a way to skip ifi_change here
        return (string_at(addressof(self), sizeof(self)) == \
            string_at(addressof(other), sizeof(other))) and \
            self.ifname == other.ifname and \
            self.flags == other.flags and \
            self.operstate == other.operstate and \
            self.linkinfo == other.linkinfo and \
            self.protinfo == other.protinfo


class KCache(RtNetlink):

    def __init__(self, pid, family=AF_UNSPEC, types=[RTM_GETLINK],
        filters={}, notifiers={}):

        RtNetlink.__init__(self, pid)
        self.cache_family = family
        self.types = types
        self.filters = filters
        self.notifiers = notifiers 

        groups = RTMGRP_LINK
        # XXX - need to collapse the clagd threads and eliminate the need 
        # for locking
        # adds and dels in the clag netlink thread's context are protected with
        # a lock (gets are not not).
        self.linkLock = threading.Lock()

        self.bind(groups, self._cbs())
        self.refill()

    def _filter(self, type, data):
        if type in self.filters:
            f = self.filters[type]
            return f(type, data)
        return False

    def _notify(self, type, data):
        if type in self.notifiers:
            f = self.notifiers[type]
            f(type, data)

    def _link_add(self, ifindex, old_iface, new_iface):
        self.linkLock.acquire()
        self.ifaces[ifindex] = new_iface
        self.linkLock.release()
        self._notify(RTM_NEWLINK, (ifindex, old_iface, new_iface))

    def _link_del(self, ifindex, old_iface):
        self.linkLock.acquire()
        del self.ifaces[ifindex]
        self.linkLock.release()
        self._notify(RTM_DELLINK, (ifindex, old_iface))

    def _link(self, nlh, ifi):

        if self._filter(nlh.nlmsg_type, ifi):
            return

        ifindex = ifi.ifi_index
        flags = ifi.ifi_flags
        old = self.ifaces.get(ifindex)

        if ifi.ifi_family == socket.AF_UNSPEC:
            which_ones = { IFLA_IFNAME : {}, IFLA_OPERSTATE : {} }
            which_ones[IFLA_LINKINFO] = { IFLA_INFO_KIND : {} }
            which_ones[IFLA_LINKINFO][IFLA_INFO_DATA] = { IFLA_VXLAN_ID : {}, IFLA_VXLAN_LOCAL : {} }
            rtas = ifi.unpack_rtas(which_ones)

            ifname = rtas.get(IFLA_IFNAME)
            operstate = rtas.get(IFLA_OPERSTATE)
            linkinfo = rtas.get(IFLA_LINKINFO)
            protinfo = {}


            if nlh.nlmsg_type == RTM_NEWLINK:
                if old:
                    protinfo = old.protinfo
                new = KCacheIface(ifi, flags, ifname, operstate, linkinfo, protinfo)
                if not (old and old.stale and old == new):
                    self._link_add(ifindex, old, new)
                # XXX - why is this happening
                if old:
                    old.stale = False
            else:
                if old:
                    self._link_del(ifindex, old)
        elif ifi.ifi_family == socket.AF_BRIDGE:
            if old:
                which_ones = { IFLA_PROTINFO : { IFLA_BRPORT_STATE : {} } }
                rtas = ifi.unpack_rtas(which_ones)

                ifname = old.ifname
                operstate = old.operstate
                linkinfo = old.linkinfo
                protinfo = rtas.get(IFLA_PROTINFO, {})

                if nlh.nlmsg_type == RTM_NEWLINK:
                    new = KCacheIface(ifi, flags, ifname, operstate, linkinfo, protinfo)
                    if not (old and old.stale and old == new):
                        self._link_add(ifindex, old, new)


    def _cbs(self):
        cbs = {}
        if RTM_GETLINK in self.types:
            cbs[RTM_NEWLINK] = self._link
            cbs[RTM_DELLINK] = self._link
        return cbs

    def _mark_stale(self):
        for ifindex, iface in self.ifaces.items():
            iface.stale = True

    def _clean_stale(self):
        stales = [(ifindex, iface) for ifindex, iface in \
            self.ifaces.items() if iface.stale]
        for ifindex, iface in stales:
            self._link_del(ifindex, iface)

    def _request_dump(self):
        for t in self.types:
            rtm = None
            if t == RTM_GETLINK:
                rtm = Rtgenmsg(self.cache_family)
            if rtm:
                token = self.request(t, NLM_F_REQUEST | NLM_F_DUMP, rtm)
                self.process_wait([token])

    def resync(self):
        self._mark_stale()
        self._request_dump()
        self._clean_stale()

    def refill(self):
        self.ifaces = {}
        self._request_dump()

    def iface_by_name(self, name):
        self.linkLock.acquire()
        match = [v for k, v in self.ifaces.items() if v.ifname == name]
        iface = match[0] if match else None
        self.linkLock.release()
        return iface

    def iface_db_copy(self):
        self.linkLock.acquire()
        iface_db = copy.deepcopy(self.ifaces)
        self.linkLock.release()
        return iface_db
