#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2014 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import clag.datasync
    import copy
    import os
    import Queue
    import subprocess
    import threading
    import json
    import clag.util as util
    import xml.etree.ElementTree as ET
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

class mdbsync(clag.datasync.datasync):
    """
    This class handles everything required to synchronize the Multicast 
    Forwarding Database, or MDB, between two CLAG peers.
    """

    # Work queue operation codes.
    _MCAST_ADD      = 1
    _MCAST_DELETE   = 2
    _MCAST_REFRESH  = 3
    _RPORT_ADD      = 4
    _RPORT_DELETE   = 5
    _RPORT_REFRESH  = 6
                  
    def __init__(self, Daemon, Intf, Parser, Log):
        """
        Constructor: Initialize data structures, threads, etc.
        """
        super(mdbsync, self).__init__()

        #
        #   Save the parameters passed in
        #
        self.Daemon = Daemon
        self.Intf   = Intf
        self.Parser = Parser
        self.Log    = Log

        #
        #   Mcast Database - The Multicast groups which are present on dual     
        #   connected interfaces (or their sub-interfaces) are tracked in this  
        #   database, which is a dictionary.  The key is a tuple containing the 
        #   multicast group, bridge, port, and VLAN ID.  The value is a list    
        #   containing the dual connected interface, the number of seconds since
        #   until the entry will be removed, and a flag indicating if the       
        #   address was added from the peer or learned on this system and a     
        #   timestamp when the address was last refreshed.                      
        #   {
        #      Multicast Group  Bridge  Port    VlanId     dualIf    Age  Remote  Updated
        #       ("234.1.1.2",   "br0",  "bond0", 101) :   [ "bond0",  42, True,       14],
        #       ("234.1.1.3",   "br0",  "bond0", 101) :   [ "bond0",  31, False,       0],
        #       ("234.1.1.4",   "br0",  "bond1", 102) :   [ "bond1", 240, True,       34]
        #   }
        #
        self.mcastDBLock    = threading.Lock()
        self.mcastDB        = {}

        #
        #   Router Port Database - The Multicast router ports which are present 
        #   on dual connected interfaces (or their sub-interfaces) are tracked  
        #   in this database, which is a dictionary.  The key is a tuple        
        #   containing the bridge and port.  The value is a list containing the 
        #   dual connected interface, the number of seconds since until the     
        #   entry will be removed, and a flag indicating if the address was     
        #   added from the peer or learned on this system and a timestamp when  
        #   the address was last refreshed.                                     
        #   {
        #        Bridge  Port         dualIf   Age  Remote  Updated
        #       ("br0",  "bond0") : [ "bond0",  42, True,       14],
        #       ("br0",  "bond0") : [ "bond0",  31, False,       0],
        #       ("br0",  "bond1") : [ "bond1", 240, True,       34]
        #   }
        #
        # This database is locked with self.mcastDBLock
        self.rportDB        = {}

        #
        #   Multicast Work Queue - This queue contains Multicast Groups and     
        #   router ports which need to be added, deleted or refreshed.  Each    
        #   entry is a tuple containing an operation, the multicast group, the  
        #   bridge, the bridge port, and a VLAN Id:                             
        #
        #       ( _MCAST_ADD, "234.1.1.2", "br0", "bond0", 101 )
        #       ( _RPORT_ADD, None,        "br0", "bond1", 0   )
        #
        self.mcastWorkQueue = Queue.Queue()
        util.queueAddNameMap(self.mcastWorkQueue, 'mcastWorkQueue')

        #
        #   Start a thread to handle updates from the peer switch
        #
        self.keep_going          = True
        self.updateThread = threading.Thread(None, self.UpdateMcastsInHwT, "UpdateMcastsInHw")
        self.updateThread.daemon = True
        self.updateThread.start()

    def __del__(self):
        """
        Destructor:
        """
        #
        #   Signal the update thread to wake up and exit
        #
        self.SignalThreads()
        self.JoinThreads()

    def UpdateMcastsInHwT(self):
        try:
            self.UpdateMcastsInHw()
        except:
            self.Daemon.DumpTraceback()

    def SignalThreads(self):
        self.keep_going = False
        try:
            self.mcastWorkQueue.put_nowait((None, None, None, None, None))
            util.queueSetMaxLen('mcastWorkQueue')
        except Queue.Full:
            pass

    def JoinThreads(self):
        self.updateThread.join()

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_MDB_SYNC, logStr)

    def WorkQueuePut(self, element):
        if self.keep_going:
            self.mcastWorkQueue.put(element)
            util.queueSetMaxLen('mcastWorkQueue')

    def __str__(self):
        """
        Return local information in human-readable form
        """
        mcastDb = []
        self.mcastDBLock.acquire()
        for (group, bridge, port, vlanId), [dualIf, age, remote, updated] in self.mcastDB.iteritems():
            if not remote:
                mcastInfo = {}
                mcastInfo['vlanId'] = vlanId
                mcastInfo['group'] = group
                mcastInfo['bridge'] = bridge
                mcastInfo['port'] = port
                mcastInfo['age'] = age
                mcastDb.append(mcastInfo)
        self.mcastDBLock.release()
        return json.dumps(mcastDb, sort_keys=True, indent=4)

    def EncodedOurStr(self):
        return self.__str__()

    def EncodedPeerStr(self):
        """
        Return peer information in human-readable form
        """
        mcastDb = []
        self.mcastDBLock.acquire()
        for (group, bridge, port, vlanId), [dualIf, age, remote, updated] in self.mcastDB.iteritems():
            if remote:
                mcastInfo = {}
                mcastInfo['vlanId'] = vlanId
                mcastInfo['group'] = group
                mcastInfo['bridge'] = bridge
                mcastInfo['port'] = port
                mcastInfo['age'] = age
                mcastDb.append(mcastInfo)
        self.mcastDBLock.release()
        return json.dumps(mcastDb, sort_keys=True, indent=4)

    def EncodedOurRportStr(self):
        """
        Return local router port information in human-readable form
        """
        mcastDb = []
        self.mcastDBLock.acquire()
        for (bridge, port), [dualIf, age, remote, updated] in self.rportDB.iteritems():
            if not remote:
                mcastInfo = {}
                mcastInfo['bridge'] = bridge
                mcastInfo['port'] = port
                mcastInfo['dualIf'] = dualIf
                mcastInfo['age'] = age
                mcastDb.append(mcastInfo)
        self.mcastDBLock.release()
        return json.dumps(mcastDb, sort_keys=True, indent=4)

    def EncodedPeerRportStr(self):
        """
        Return local router port information in human-readable form
        """
        mcastDb = []
        self.mcastDBLock.acquire()
        for (bridge, port), [dualIf, age, remote, updated] in self.rportDB.iteritems():
            if remote:
                mcastInfo = {}
                mcastInfo['bridge'] = bridge
                mcastInfo['port'] = port
                mcastInfo['dualIf'] = dualIf
                mcastInfo['age'] = age
                mcastDb.append(mcastInfo)
        self.mcastDBLock.release()
        return json.dumps(mcastDb, sort_keys=True, indent=4)

    def CollectLocalInfo(self):
        """
        This function is called on a periodic interval to gather up any 
        information on the local switch which needs to be collected prior to
        sending that information to the peer switch.
        """

        # Get the current list of dual connected interfaces
        dualIfs = self.Intf.GetDualIfs()

        # Get a mapping of VLAN sub-interfaces to their master interfaces and VLAN IDs
        vlanMap = self.Intf.GetSubIfsVlanMap()

        # For each bridge with dual connected interfaces, get the multicast membership
        # interval and querier interval.
        bridgeMemIntval     = None
        bridgeQuerierIntval = None

        # Get the current multicst address table
        try:
            mcastTableStr = subprocess.check_output(["/sbin/bridge", "-s", "-d", "mdb", "show"])
        except subprocess.CalledProcessError:
            return
        now = int(os.times()[4])

        # Make a set of all multicast groups we knew about from last poll and   
        # sync.  As we get each one from the kernel it will be removed from this
        # set, leaving those multicast groups which have been removed since last
        # poll and sync.                                                        
        self.mcastDBLock.acquire()
        removedMcasts = set(self.mcastDB)
        removedRports = set(self.rportDB)

        for mcastLineStr in mcastTableStr.splitlines():
            # Is this a dynamic group (no "permanent" in it)
            if "permanent" not in mcastLineStr:
                self.log_debug("Processing entry: %s" % (mcastLineStr,))
                mcastLineList = mcastLineStr.split()
                if mcastLineList[0] == "dev" and len(mcastLineList) >= 2:
                    k_bridge = mcastLineList[1]

                    try: # to get the port
                        portIdx = mcastLineList.index('port')
                        k_port = mcastLineList[portIdx+1]
                    except (ValueError, IndexError):
                        continue

                    try: # to get the group
                        grpIdx = mcastLineList.index('grp')
                        k_group = mcastLineList[grpIdx+1]
                    except (ValueError, IndexError):
                        continue

                    try: # to get the ttl
                        k_age = int(float(mcastLineList[-1]))
                    except ValueError:
                        continue

                    try: # to get the vlan id
                        vidIdx = mcastLineList.index('vid')
                        k_vlanId = int(mcastLineList[vidIdx+1])
                        k_dualIf = k_port
                    except (ValueError, IndexError):
                        (k_dualIf, k_vlanId) = vlanMap.get(k_port, (k_port, None))

                    if k_dualIf in dualIfs:
                        mcastKey = (k_group, k_bridge, k_port, k_vlanId)
                        # Is this a new group we don't know anything about yet?
                        if mcastKey not in self.mcastDB:
                            self.mcastDB[mcastKey] = [k_dualIf, k_age, False, 0]
                            self.log_debug("New local multicast group %s bridge %s port %s vlanId %s dualIf %s age %s" % (k_group, k_bridge, k_port, k_vlanId, k_dualIf, k_age))
                        else:
                            [c_dualIf, c_age, c_remote, c_updated] = self.mcastDB[mcastKey]
                            removedMcasts.discard(mcastKey)
                            # Is the multicast group owned by the peer?
                            if c_remote:
                                # Have we seen the address more recently than the peer?
                                if bridgeMemIntval is None:
                                    bridgeMemIntval = self.GetBridgeMembershipIntervals(dualIfs)
                                max_age = bridgeMemIntval.get(k_port, 260)
                                if (k_age - 15) >= (max_age - (now - c_updated)):
                                    # Yes. Make the address local
                                    self.mcastDB[mcastKey] = [k_dualIf, k_age, False, 0]
                                    self.log_debug("Moving to local: multicast group %s bridge %s port %s vlanId %s dualIf %s age %s" % (k_group, k_bridge, k_port, k_vlanId, k_dualIf, k_age))
                            else:
                                # Multicast group is owned by this switch. Update/add Mcast entry
                                self.mcastDB[mcastKey] = [k_dualIf, k_age, False, 0]
                                #self.log_debug("Updating age: multicast group %s bridge %s port %s vlanId %s dualIf %s age %s" % (k_group, k_bridge, k_port, k_vlanId, k_dualIf, k_age))
                elif mcastLineList[0] == "router" and len(mcastLineList) >= 7:
                    # router ports on br0: bond1  254.40 temp
                    k_bridge = mcastLineList[3]
                    k_bridge = k_bridge[:-1] if k_bridge[-1:] == ":" else k_bridge
                    k_port   = mcastLineList[4]
                    k_age    = int(float(mcastLineList[5]))
                    k_dualIf = vlanMap.get(k_port, (k_port, None))[0] if self.Intf.isSubIfName(k_port) else k_port
                    if k_dualIf in dualIfs:
                        rportKey = (k_bridge, k_port)
                        # Is this a new router port we don't know anything about yet?
                        if rportKey not in self.rportDB:
                            self.rportDB[rportKey] = [k_dualIf, k_age, False, 0]
                            self.log_debug("New local router port %s bridge %s dualIf %s age %s" % (k_port, k_bridge, k_dualIf, k_age))
                        else:
                            [c_dualIf, c_age, c_remote, c_updated] = self.rportDB[rportKey]
                            removedRports.discard(rportKey)
                            # Is the router port owned by the peer?
                            if c_remote:
                                # Have we seen the router port more recently than the peer?
                                if bridgeQuerierIntval is None:
                                    bridgeQuerierIntval = self.GetBridgeQuerierIntervals(dualIfs)
                                max_age = bridgeQuerierIntval.get(k_port, 255)
                                if (k_age - 15) >= (max_age - (now - c_updated)):
                                    # Yes. Make the router port local
                                    self.rportDB[rportKey] = [k_dualIf, k_age, False, 0]
                                    self.log_debug("Moving to local: router port %s bridge %s dualIf %s age %s" % (k_port, k_bridge, k_dualIf, k_age))
                            else:
                                # Router port is owned by this switch. Update/add entry
                                self.rportDB[rportKey] = [k_dualIf, k_age, False, 0]
                                #self.log_debug("Updating age: router port %s bridge %s dualIf %s age %s" % (k_port, k_bridge, k_dualIf, k_age))

        # Delete all entries in mcastDB which are no longer in HW
        for r in removedMcasts:
            self.log_debug("Removed from HW: multicast group %s bridge %s port %s vlanId %s dualIf %s age %s" % (r[0], r[1], r[2], r[3], self.mcastDB.get(r,[None])[0], self.mcastDB.get(r,[None, None])[1]))
            self.mcastDB.pop(r, None)

        # Delete all entries in rportDB which are no longer in HW
        for r in removedRports:
            self.log_debug("Removed from HW: router port %s bridge %s dualIf %s age %s" % (r[1], r[0], self.rportDB.get(r,[None])[0], self.rportDB.get(r,[None, None])[1]))
            self.rportDB.pop(r, None)

        self.mcastDBLock.release()

    def GetBridgeMembershipIntervals(self, intfs):
        membershipIntervals = {}
        masterVlanMap = self.Intf.GetMasterIfVlanMap()
        for intf in intfs:
            age = self.Intf.GetBridgeMemInterval(self.Intf.GetBridgeOfMember(intf))
            if age is not None:
                membershipIntervals[intf] = age
            for subIf in masterVlanMap.get(intf, {}).values():
                age = self.Intf.GetBridgeMemInterval(self.Intf.GetBridgeOfMember(subIf))
                if age is not None:
                    membershipIntervals[subIf] = age
        return membershipIntervals

    def GetBridgeQuerierIntervals(self, intfs):
        querierIntervals = {}
        masterVlanMap = self.Intf.GetMasterIfVlanMap()
        for intf in intfs:
            age = self.Intf.GetBridgeQuerierInterval(self.Intf.GetBridgeOfMember(intf))
            if age is not None:
                querierIntervals[intf] = age
            for subIf in masterVlanMap.get(intf, {}).values():
                age = self.Intf.GetBridgeQuerierInterval(self.Intf.GetBridgeOfMember(subIf))
                if age is not None:
                    querierIntervals[subIf] = age
        return querierIntervals

    def EncodeXmlInfo(self, parent):
        """
        Encodes the local information as XML.

        parent - An xml.etree.ElementTree.Element instance which is the parent
            element under which all of the information should be encoded.

        <multicast>
            <group dualIf="bond2" vlan="100" dev="br0" age="259">232.10.10.10</group>
            <group dualIf="bond3" vlan="101" dev="br1" age="13">232.10.10.11</group>
            <group dualIf="bond4" dev="br2" age="125">232.10.10.12</group>
            <router_port dualIf="bond4" dev="br2" age="125"/>
            <router_port dualIf="bond0" dev="br0" age="14"/>
        </multicast>
        """
        self.mcastDBLock.acquire()
        mcastDB = copy.deepcopy(self.mcastDB)
        rportDB = copy.deepcopy(self.rportDB)
        self.mcastDBLock.release()
        mcast = ET.SubElement(parent, "multicast")
        for (group, bridge, port, vlanId), [dualIf, age, remote, updated] in mcastDB.items():
            if not remote:
                mcastAttrib = { "dualIf" : dualIf, "dev" : bridge, "age" : str(age) }
                if vlanId is not None:
                    mcastAttrib["vlan"] = str(vlanId)
                ET.SubElement(mcast, "group", mcastAttrib).text = group
        for (bridge, port), [dualIf, age, remote, updated] in rportDB.items():
            if not remote:
                rportAttrib = { "dualIf" : dualIf, "dev" : bridge, "age" : str(age) }
                ET.SubElement(mcast, "router_port", rportAttrib)

    def DecodeXmlInfo(self, root):
        """
        Decodes XML-encoded information sent from the peer.

        root - The root element of the XML tree
        """
        if root.find('multicast') is not None:
            # Get the current list of dual connected hosts
            dualIfs    = self.Intf.GetDualIfs()
            revDualIfs = { v : k for k , v in dualIfs.items() }

            # Get a mapping of master interfaces to sub-interface names and VLAN IDs
            vlanMap = self.Intf.GetMasterIfVlanMap()

        # For each bridge with dual connected interfaces, get the multicast membership interval.
        bridgeMemIntval     = None
        bridgeQuerierIntval = None

        self.mcastDBLock.acquire()
        now = int(os.times()[4])
        for mcast in root.findall('multicast'):
            for group_element in mcast.findall('group'):
                r_peerDualIf = group_element.get('dualIf')
                r_dualIf = revDualIfs.get(r_peerDualIf)
                r_age = int(group_element.get('age'))
                r_group = group_element.text
                r_vlanId = group_element.get('vlan')
                if r_vlanId is not None:
                    r_vlanId = int(r_vlanId)

                # Convert the r_dualIf, r_vlanId into a bridge, port, vlanId
                port = vlanMap.get(r_dualIf, {}).get(r_vlanId, r_dualIf)
                bridge = self.Intf.GetBridgeOfMember(port)
                if bridge is not None:
                    mcastKey = (r_group, bridge, port, r_vlanId)
                    if mcastKey not in self.mcastDB:
                        self.mcastDB[mcastKey] = [r_dualIf, r_age, True, now]
                        self.WorkQueuePut((mdbsync._MCAST_ADD, r_group, bridge, port, r_vlanId))
                        self.log_debug("Added new remote: Multicast group %s bridge %s port %s vlanId %s age %s" % (r_group, bridge, port, r_vlanId, r_age))
                    else:
                        [c_dualIf, c_age, c_remote, c_updated] = self.mcastDB[mcastKey]
                        if c_remote:
                            if bridgeMemIntval is None:
                                bridgeMemIntval = self.GetBridgeMembershipIntervals(dualIfs)
                            max_age = bridgeMemIntval.get(port, 260)
                            if (now - c_updated) > (max_age * 2/3):
                                c_updated = now
                                self.WorkQueuePut((mdbsync._MCAST_REFRESH, r_group, bridge, port, r_vlanId))
                                self.log_debug("Refreshing remote: Multicast group %s bridge %s port %s vlanId %s age %s" % (r_group, bridge, port, r_vlanId, r_age))
                            self.mcastDB[mcastKey] = [r_dualIf, r_age, True, c_updated]
                        else:
                            if r_age > c_age:
                                self.mcastDB[mcastKey] = [r_dualIf, r_age, True, now]
                                self.WorkQueuePut((mdbsync._MCAST_REFRESH, r_group, bridge, port, r_vlanId))
                                self.log_debug("Moving local to remote: Multicast group %s bridge %s port %s vlanId %s age %s" % (r_group, bridge, port, r_vlanId, r_age))
            for rport_element in mcast.findall('router_port'):
                r_peerDualIf = rport_element.get('dualIf')
                r_dualIf = revDualIfs.get(r_peerDualIf)
                r_age = int(rport_element.get('age'))

                # Convert the r_dualIf into a bridge
                bridge = self.Intf.GetBridgeOfMember(r_dualIf)
                port = r_dualIf
                if bridge is not None:
                    rportKey = (bridge, port)
                    if rportKey not in self.rportDB:
                        self.rportDB[rportKey] = [r_dualIf, r_age, True, now]
                        self.WorkQueuePut((mdbsync._RPORT_ADD, None, bridge, port, 0))
                        self.log_debug("Added new remote: Multicast router port %s bridge %s age %s" % (port, bridge, r_age))
                    else:
                        [c_dualIf, c_age, c_remote, c_updated] = self.rportDB[rportKey]
                        if c_remote:
                            if bridgeQuerierIntval is None:
                                bridgeQuerierIntval = self.GetBridgeQuerierIntervals(dualIfs)
                            max_age = bridgeQuerierIntval.get(port, 255)
                            if (now - c_updated) > (max_age * 2/3):
                                c_updated = now
                                self.WorkQueuePut((mdbsync._RPORT_REFRESH, None, bridge, port, 0))
                                self.log_debug("Refreshing remote: Multicast router port %s bridge %s age %s" % (port, bridge, r_age))
                            self.rportDB[rportKey] = [r_dualIf, r_age, True, c_updated]
                        else:
                            if r_age > c_age:
                                self.rportDB[rportKey] = [r_dualIf, r_age, True, now]
                                self.WorkQueuePut((mdbsync._RPORT_REFRESH, None, bridge, port, 0))
                                self.log_debug("Moving local to remote: Multicast router port %s bridge %s age %s" % (port, bridge, r_age))
        self.mcastDBLock.release()

    def DualIfChange(self, dualIf, enable):
        """
        Called when an interface becomes or is no longer dual connected
        """
        pass

    def ClearPeerInfo(self):
        """
        We don't clear multicast group information when the peer link goes down
        or when we exit clagd. So this is a NOP.
        """
        pass

    def ClearLocalInfo(self):
        """
        Removes multicast groups which are owned by this switch.
        """
        self.mcastDBLock.acquire()
        for (group, bridge, port, vlanId), [dualIf, age, remote, updated] in mcastDB.items():
            if not remote:
                self.WorkQueuePut((mdbsync._MCAST_DELETE, group, bridge, port, vlanId))
                del self.mcastDB[(group, bridge, port, vlanId)]
        for (bridge, port), [dualIf, age, remote, updated] in rportDB.items():
            if not remote:
                self.WorkQueuePut((mdbsync._RPORT_DELETE, None, bridge, port, 0))
                del self.rportDB[(bridge, port)]
        self.mcastDBLock.release()

    def UpdateMcastsInHw(self):
        '''
        This function is an independent thread of the script. 
        '''
        mdb_cmd = {
            mdbsync._MCAST_ADD     : ( "replace", "added to"     ),
            mdbsync._MCAST_REFRESH : ( "replace", "refreshed on" ),
            mdbsync._MCAST_DELETE  : ( "del",     "removed from" ),
            mdbsync._RPORT_ADD     : ( "3",       "added to"     ),
            mdbsync._RPORT_REFRESH : ( "3",       "refreshed on" ),
            mdbsync._RPORT_DELETE  : ( "1",       "removed from" )
        }

        self.log_debug("Beginning execution of the thread UpdateMcastsInHw")

        while self.keep_going:
            # Wait for a change in multicast group data base received from the peer
            self.log_debug("Waiting for an event before continuing execution of the thread UpdateMcastsInHw")
            queueEmpty = False
            try:
                (op, group, bridge, port, vlanId) = self.mcastWorkQueue.get(True, self.Parser.getPeriodicRun())
            except Queue.Empty:
                queueEmpty = True

            if self.keep_going:
                params = []
                while not queueEmpty and self.keep_going:
                    (mdb_op, mdb_str) = mdb_cmd.get(op, (None, None))
                    if op in [mdbsync._MCAST_ADD, mdbsync._MCAST_REFRESH, mdbsync._MCAST_DELETE]:
                        if self.Intf.isSubIfName(port) or vlanId is None:
                            params += [(mdb_op, bridge, port, group, "")]
                            self.Log.log_verbose("Multicast group %s has been %s device %s port %s." % (group, mdb_str, bridge, port))
                        else:
                            vlanStr = " vid %s" % (vlanId,)
                            params += [(mdb_op, bridge, port, group, vlanStr)]
                            self.Log.log_verbose("Multicast group %s has been %s device %s port %s vid %s." % (group, mdb_str, bridge, port, vlanId))
                    elif op in [mdbsync._RPORT_ADD, mdbsync._RPORT_REFRESH, mdbsync._RPORT_DELETE]:
                        DEVNULL = open(os.devnull, 'wb')
                        cmd = ["/sbin/brctl", "setportmcrouter", bridge, port, mdb_op]
                        self.Log.log_verbose("Multicast router port %s has been %s bridge %s." % (port, mdb_str, bridge))
                        subprocess.call(cmd, stdout=DEVNULL, stderr=DEVNULL)
                    else:
                        self.Log.log_error("Unknown op code in work queue: %s" % (op,))

                    self.mcastWorkQueue.task_done()
                    queueEmpty = self.mcastWorkQueue.empty()
                    if not queueEmpty:
                        try:
                            (op, group, bridge, port, vlanId) = self.mcastWorkQueue.get(True, self.Parser.getPeriodicRun())
                        except Queue.Empty:
                            self.log_debug("Queue was empty after checking if it was non-empty")
                            queueEmpty = True

                if not queueEmpty:
                    self.mcastWorkQueue.task_done()

                self.Intf.ExecuteBatch(["/sbin/bridge", "-force", "-batch"],"mdb %s dev %s port %s grp %s temp%s", params)

        # Burn all the remaining entries in the queue
        while not self.mcastWorkQueue.empty():
            self.mcastWorkQueue.get()
            self.mcastWorkQueue.task_done()

        self.log_debug("Finished execution of the thread UpdateMcastsInHw")

