#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2014 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import clag.clagthread
    import clag.datasync
    import copy
    import threading
    import json
    import xml.etree.ElementTree as ET
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

class vlansync(clag.datasync.datasync):
    """
    This class handles everything required to send and receive the VLAN information
    between two CLAG peers.
    """

    def __init__(self, Intf, Parser, Log, peerSendQueue):
        """
        Constructor: Initialize data structures, threads, etc.
        """
        super(vlansync, self).__init__()

        #
        #   Save the parameters passed in
        #
        self.Intf   = Intf
        self.Parser = Parser
        self.Log    = Log
        self.peerSendQueue = peerSendQueue

        #
        #   VLAN Database - There are two databases, one for our switch and one
        #   for the peer switch. The structure of each data base is the same: 
        #   A dictionary which contains the dual
        #   connected interface names as the keys, and a list of 128 32-bit 
        #   numbers. The list contains 4096 bits, one for each possible VLAN ID. 
        #   The first element in the list represents VLAN IDs 31-0 with MSB
        #   VLAN ID 31 and the LSB VLAN ID 0. 
        #   {
        #       "bond3" : [ 0x00000001, 0x00010013, ... ]
        #       "bond2" : [ 0x00010F00, 0x00A00700, ... ]
        #   }
        #
        self.ourVlanDBLock  = threading.Lock()
        self.ourVlanDB      = {}
        self.peerVlanDBLock = threading.Lock()
        self.peerVlanDB     = {}
        self.peerReplyEvent = clag.clagthread.Event()

    def __del__(self):
        """
        Destructor:
        """
        pass

    def SignalThreads(self):
        self.peerReplyEvent.set()

    def JoinThreads(self):
        pass

    def __str__(self):
        """
        Return local information in human-readable form
        """
        vlanDb = self.GetOurVlanDB()
        return json.dumps(vlanDb, sort_keys=True, indent=4)

    def EncodedOurStr(self):
        return self.__str__()

    def EncodedPeerStr(self):
        """
        Return peer information in human-readable form
        """
        vlanDb = self.GetPeerVlanDB()
        return json.dumps(vlanDb, sort_keys=True, indent=4)

    def GetOurVlanDB(self):
        """
        Return a copy of the current VLAN data base on this switch.
        """
        self.FillOurVlanDB()
        self.ourVlanDBLock.acquire()
        vlanDB = copy.deepcopy(self.ourVlanDB)
        self.ourVlanDBLock.release()
        return vlanDB

    def GetPeerVlanDB(self):
        """
        Return a copy of the current VLAN data base on the peer switch.
        """
        self.FillPeerVlanDB()
        self.peerVlanDBLock.acquire()
        vlanDB = copy.deepcopy(self.peerVlanDB)
        self.peerVlanDBLock.release()
        return vlanDB

    def SetOurVlanDB(self, vlanDB):
        """
        Set the current value of the VLAN data base on this switch.
        """
        self.ourVlanDBLock.acquire()
        if self.ourVlanDB != vlanDB:
            self.ourVlanDB = copy.deepcopy(vlanDB)
        self.ourVlanDBLock.release()

    def SetPeerVlanDB(self, vlanDB):
        """
        Return a copy of the current VLAN data base on the peer switch.
        """
        self.peerVlanDBLock.acquire()
        if self.peerVlanDB != vlanDB:
            self.peerVlanDB = copy.deepcopy(vlanDB)
        self.peerVlanDBLock.release()

    def CollectLocalInfo(self):
        """
        This function is called on a periodic basis to gather up the 
        information on the local switch which needs to be collected prior to
        sending that information to the peer switch.

        Since the VLAN information is only collected on an "as needed" basis,
        this function does nothing.
        """
        pass

    def FillOurVlanDB(self):
        """
        This function is called to fill in the local VLAN database.
        """
        # Get the current list of dual connected hosts
        dualIfs = self.Intf.GetDualIfs()
        vlanMap = self.Intf.GetMasterIfVlanMap()

        vlanDB = {}
        for dualIf in dualIfs:
            # Depending on the bridge model, the VLAN IDs are obtained in one of 
            # two ways. First, try the new bridge model:
            vlanDB[dualIf] = self.Intf.GetIfVlanMap(dualIf)
            if len(vlanDB[dualIf]) == 0:
                vlanDB[dualIf] = [0x00000000] * 128
                for vlanId in vlanMap.get(dualIf,{}):
                    vlanDB[dualIf][vlanId // 32] |= 1 << (vlanId % 32)

        self.SetOurVlanDB(vlanDB)

    def EncodeXmlInfo(self, parent):
        '''
        This module does not send periodic information to the peer. All
        communication to the peer is event driven.
        '''
        pass

    def FillPeerVlanDB(self):
        '''
        In order to fill the peer's VLAN database, we send a request to the peer
        and wait for him to reply.
        '''
        self.peerReplyEvent.clear()
        self.SendRequestToPeer()
        timeout = not self.peerReplyEvent.wait(6)
        if timeout:
            self.ClearPeerInfo()

    def SendRequestToPeer(self):
        '''
        This function sends a request to the peer to return its VLAN information.
        '''
        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        vlans = ET.SubElement(root, "vlan-request")
        xmlStr = ET.tostring(root)
        self.peerSendQueue.put(xmlStr)

    def SendInfoToPeer(self):
        """
        Encodes the local information as XML.

        parent - An xml.etree.ElementTree.Element instance which is the parent
            element under which all of the information should be encoded.

        <vlans>
            <vlanmap dualIf="bond4">
                0x00000100 0x02D04002 0xD3000145 ...
            </vlanmap>
            <vlanmap dualIf="bond5">
                0x00000100 0x02D04002 0xD3000145 ...
            </vlanmap>
        </vlans>
        """
        vlanDB = self.GetOurVlanDB()
        root = ET.Element("clag_data", {'version' : self.Parser._ClagDataVersion})
        vlans = ET.SubElement(root, "vlans")
        for bond in vlanDB:
            vlanMapStr = ""
            for vlanMap in vlanDB[bond]:
                vlanMapStr += "0x%08x " % (vlanMap,)
            vlanAttrib = { "dualIf" : bond }
            ET.SubElement(vlans, "vlanmap", vlanAttrib).text = vlanMapStr
        xmlStr = ET.tostring(root)
        self.peerSendQueue.put(xmlStr)

    def DecodeXmlInfo(self, root):
        """
        Decodes XML-encoded information sent from the peer.

        root - The root element of the XML tree
        """
        vlans = root.find('vlans')
        if vlans is not None:
            vlanDB = {}
            for vlanmap in vlans.findall('vlanmap'):
                dualIf = vlanmap.get('dualIf')
                vlanDB[dualIf] = [ int(i,0) for i in vlanmap.text.split() ]
            self.SetPeerVlanDB(vlanDB)
            self.peerReplyEvent.set()
            if not self.Parser.isPeerVlanSyncOptimized():
                self.SendInfoToPeer()
        vlanrequest = root.find('vlan-request')
        if vlanrequest is not None:
            self.SendInfoToPeer()

    def DualIfChange(self, dualIf, enable):
        """
        Called when an interface becomes or is no longer dual connected
        """
        pass

    def ClearPeerInfo(self):
        """
        Removes information which has been obtained from the peer.
        """
        self.SetPeerVlanDB({})

    def ClearLocalInfo(self):
        """
        Removes information which has been obtained locally.
        """
        self.SetOurVlanDB({})
