#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2014 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import clag.clagthread
    import clag.datasync
    import copy
    import os
    import re
    import socket
    import subprocess
    import threading
    import xml.etree.ElementTree as ET
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

class vxlansync(clag.datasync.datasync):
    """
    This class handles everything required to synchronize the VxLAN information
    between two CLAG peers.
    """

    _InterfacePath = "/sys/class/net/"
    _VXLAN_DB_FLAGS_NONE    = 0
    _VXLAN_DB_FLAGS_DUAL    = (1 << 0)

    def __init__(self, Daemon, Intf, Parser, Log, peerSendQueue):
        """
        Constructor: Initialize data structures, threads, etc.
        """
        super(vxlansync, self).__init__()

        #
        #   Save the parameters passed in
        #
        self.Daemon = Daemon
        self.Intf   = Intf
        self.Parser = Parser
        self.Log    = Log
        self.peerSendQueue = peerSendQueue
        self.Cmd    = None
        self.DataSyncs = []

        #
        #   VxLAN Database - There are two VxLAN databases, one for our switch and one for
        #   the peer switch. Each has the same structure. The database is a dictionary
        #   whose keys are the interface names of all vxlan devices in the system. The data
        #   is a tuple with four elements. The first element is the vxlan id used by this
        #   vxlan device. The second is the original local IP address assigned to this
        #   vxlan device. The third is the operatation state of the vxlan. The fourth is
        #   a flag which indicates if the device is dual-connected or not.
        #   {
        #       "vx-14567102" : ( 14567102, "27.0.0.11", "up", 1 ),
        #       "vx-14567103" : ( 14567103, "27.0.0.11", "down", 0 )
        #   }
        #
        self.ourVxLanDBLock      = threading.Lock()
        self.ourVxLanDB          = {}
        self.peerVxLanDBLock     = threading.Lock()
        self.peerVxLanDB         = {}
        self.currDualIfs         = set()
        self.newRole             = False
        self.waitForUp           = set()

        #
        #   Start a thread to handle updates from the peer switch
        #
        self.keep_going         = True
        self.stopEvent          = clag.clagthread.Event()
        self.newVxLanDataEvent  = clag.clagthread.Event()

        self.updateThread = threading.Thread(None, self.UpdateVxLanFromPeerT, "UpdateVxLanFromPeer")
        self.updateThread.daemon = True
        self.updateThread.start()

        self.configThread = threading.Thread(None, self.UpdateVxLanConfigT, "UpdateVxLanConfig")
        self.configThread.daemon = True
        self.configThread.start()

    def __del__(self):
        """
        Destructor:
        """
        #
        #   Signal the update thread to wake up and exit
        #
        self.SignalThreads()
        self.JoinThreads()

    def UpdateVxLanFromPeerT(self):
        try:
            self.UpdateVxLanFromPeer()
        except:
            self.Daemon.DumpTraceback()

    def UpdateVxLanConfigT(self):
        try:
            self.UpdateVxLanConfig()
        except:
            self.Daemon.DumpTraceback()

    def SignalThreads(self):
        self.keep_going = False
        self.stopEvent.set()
        self.newVxLanDataEvent.set()

    def JoinThreads(self):
        self.updateThread.join()
        self.configThread.join()

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_VXLAN_SYNC, logStr)

    def OperStToDispStr(self, operSt):
        if operSt == "dormant" or operSt == "up":
            operStr = operSt
        else:
            operStr = "down"
        return operStr

    def FlagsToDispStr(self, flags):
        if flags:
            if (flags & self._VXLAN_DB_FLAGS_DUAL):
                flagStr = "D"
        else:
            flagStr = "-"
        return flagStr

    def __str__(self):
        """
        Return local information in human-readable form
        """
        retStr = ""
        self.ourVxLanDBLock.acquire()
        for vxlan in self.ourVxLanDB:
            vxLanInfo = self.ourVxLanDB[vxlan]
            operStr = self.OperStToDispStr(vxLanInfo[2])
            flagStr = self.FlagsToDispStr(vxLanInfo[3])
            retStr += " ".join([str(vxlan), str(vxLanInfo[0]), str(vxLanInfo[1]), operStr, flagStr]) + "\n"
        self.ourVxLanDBLock.release()
        return retStr

    def EncodedOurStr(self):
        return self.__str__()

    def EncodedPeerStr(self):
        """
        Return peer information in human-readable form
        """
        retStr = ""
        self.peerVxLanDBLock.acquire()
        for vxlan in self.peerVxLanDB:
            vxLanInfo = self.peerVxLanDB[vxlan]
            operStr = self.OperStToDispStr(vxLanInfo[2])
            flagStr = self.FlagsToDispStr(vxLanInfo[3])
            retStr += " ".join([str(vxlan), str(vxLanInfo[0]), str(vxLanInfo[1]), operStr, flagStr]) + "\n"
        self.peerVxLanDBLock.release()
        return retStr

    def GetOurVxLanDB(self):
        """
        Return a copy of the current VxLAN data base on this switch.
        """
        self.ourVxLanDBLock.acquire()
        vxLanDB = copy.deepcopy(self.ourVxLanDB)
        self.ourVxLanDBLock.release()
        return vxLanDB

    def GetOurVxLanInfo(self, vxlan):
        """
        Return a copy of the current VxLAN data for a specific vxlan
        """
        self.ourVxLanDBLock.acquire()
        if vxlan in self.ourVxLanDB:
            vxLanInfo = copy.deepcopy(self.ourVxLanDB[vxlan])
        else:
            vxLanInfo = None
        self.ourVxLanDBLock.release()
        return vxLanInfo

    def GetOurVxLanInfoByVni(self, vni):
        """
        Return a copy of the our VxLAN data for a specific VNI
        """
        vxLanInfo = None
        self.ourVxLanDBLock.acquire()
        for vxlan in self.ourVxLanDB:
            if vni == self.ourVxLanDB[vxlan][0]:
                vxLanInfo = (vxlan, copy.deepcopy(self.ourVxLanDB[vxlan]))
                break
        self.ourVxLanDBLock.release()
        return vxLanInfo

    def GetPeerVxLanDB(self):
        """
        Return a copy of the current VxLAN data base on the peer switch.
        """
        self.peerVxLanDBLock.acquire()
        vxLanDB = copy.deepcopy(self.peerVxLanDB)
        self.peerVxLanDBLock.release()
        return vxLanDB

    def GetPeerVxLanInfo(self, vxlan):
        """
        Return a copy of the peer VxLAN data for a specific vxlan
        """
        self.peerVxLanDBLock.acquire()
        if vxlan in self.peerVxLanDB:
            vxLanInfo = copy.deepcopy(self.peerVxLanDB[vxlan])
        else:
            vxLanInfo = None
        self.peerVxLanDBLock.release()
        return vxLanInfo

    def GetPeerVxLanInfoByVni(self, vni):
        """
        Return a copy of the peer VxLAN data for a specific VNI
        """
        vxLanInfo = None
        self.peerVxLanDBLock.acquire()
        for vxlan in self.peerVxLanDB:
            if vni == self.peerVxLanDB[vxlan][0]:
                vxLanInfo = copy.deepcopy(self.peerVxLanDB[vxlan])
                break
        self.peerVxLanDBLock.release()
        return vxLanInfo

    def SetOurVxLanDB(self, vxLanDB):
        """
        Set the current value of the VxLAN data base on this switch.
        """
        self.ourVxLanDBLock.acquire()
        if self.ourVxLanDB != vxLanDB:
            self.ourVxLanDB = copy.deepcopy(vxLanDB)
            self.newVxLanDataEvent.set()
        self.ourVxLanDBLock.release()

    def AddVxLanInterface(self, ifname, vni, localip, operstate, brportst):
        self.ourVxLanDBLock.acquire()
        if ifname not in self.ourVxLanDB:
            flags = self._VXLAN_DB_FLAGS_DUAL if self.Intf.isDualIf(ifname) else self._VXLAN_DB_FLAGS_NONE
            self.ourVxLanDB[ifname] = (vni, localip, operstate, flags)
            self.Parser.addClagVxLan(ifname)
            self.newVxLanDataEvent.set()
            self.log_debug("Adding new vxlan to our data base: %s, vni=%d, localip=%s, op=%s" % (ifname, vni, localip, operstate))
        elif self.ourVxLanDB[ifname][0] != vni or self.ourVxLanDB[ifname][2] != operstate:
            self.ourVxLanDB[ifname] = (vni, self.ourVxLanDB[ifname][1], operstate, self.ourVxLanDB[ifname][3])
            self.newVxLanDataEvent.set()
            self.log_debug("Updating vxlan in our data base: %s, vni=%d, op=%s, flags=0x%02x" % (ifname, vni, operstate, self.ourVxLanDB[ifname][3]))
        anycastIp = self.Parser.args.vxlanAnycast
        if ifname in self.waitForUp and anycastIp and brportst == 3:
            self.log_debug("Adding anycastIp to the loopback interface because %s is forwarding." % (ifname,))
            DEVNULL = open(os.devnull, 'wb')
            # Add the anycast IP on the loopback interface
            cmd = [ "/sbin/ip", "addr", "add", "dev", "lo", str(anycastIp) + "/32", "scope", "global" ]
            self.log_debug("Executing: %s " % (cmd,))
            subprocess.call(cmd, stdout=DEVNULL, stderr=DEVNULL)
            self.Intf.NetlinkFilter(set(), {socket.AF_BRIDGE})
            self.waitForUp = set()
        self.ourVxLanDBLock.release()

    def DelVxLanInterface(self, ifname):
        self.ourVxLanDBLock.acquire()
        if ifname in self.ourVxLanDB:
            del self.ourVxLanDB[ifname]
            self.Parser.delClagVxLan(ifname)
            self.newVxLanDataEvent.set()
            self.log_debug("Deleting vxlan from our data base: %s" % (ifname,))
        self.ourVxLanDBLock.release()

    def SetVxLanFlags(self, vxlanName, flags):
        updateDone = False
        self.ourVxLanDBLock.acquire()
        if vxlanName in self.ourVxLanDB and not (self.ourVxLanDB[vxlanName][3] & flags):
            oldInfo = copy.deepcopy(self.ourVxLanDB[vxlanName])
            vxLanInfo = (oldInfo[0], oldInfo[1], oldInfo[2], oldInfo[3]|flags)
            self.ourVxLanDB[vxlanName] = vxLanInfo
            updateDone = True
        self.ourVxLanDBLock.release()
        if updateDone:
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, '%s dual-connect update sent to peer' % vxlanName)
            self.SendVxLanAdd(vxlanName)
        return updateDone

    def ClearVxLanFlags(self, vxlanName, flags):
        self.ourVxLanDBLock.acquire()
        if vxlanName in self.ourVxLanDB and (self.ourVxLanDB[vxlanName][3] & flags):
            oldInfo = copy.deepcopy(self.ourVxLanDB[vxlanName])
            vxLanInfo = (oldInfo[0], oldInfo[1], oldInfo[2], oldInfo[3]&~flags)
            self.ourVxLanDB[vxlanName] = vxLanInfo
        self.ourVxLanDBLock.release()

    def SetPeerVxLanDB(self, vxLanDB):
        """
        Return a copy of the current VxLAN data base on the peer switch.
        """
        self.peerVxLanDBLock.acquire()
        if self.peerVxLanDB != vxLanDB:
            self.peerVxLanDB = copy.deepcopy(vxLanDB)
            self.newVxLanDataEvent.set()
        self.peerVxLanDBLock.release()

    def AddVxLanToPeerVxLanDB(self, vxlanName, vxlanInfo):
        """
        Adds/updates a single element to the peer vxLan db
        """
        updateDone = False
        self.peerVxLanDBLock.acquire()
        if vxlanName not in self.peerVxLanDB or self.peerVxLanDB[vxlanName] != vxlanInfo:
            if vxlanName not in self.peerVxLanDB or \
               self.peerVxLanDB[vxlanName][0] != vxlanInfo[0] or \
               self.peerVxLanDB[vxlanName][1] != vxlanInfo[1] or \
               self.peerVxLanDB[vxlanName][2] != vxlanInfo[2]:
                updateDone = True
            self.peerVxLanDB[vxlanName] = copy.deepcopy(vxlanInfo)
            # XXX - need a way to avoid complete table analysis eveytime
            # a single vxlan is updated
        self.peerVxLanDBLock.release()
        if updateDone:
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'peer vxlan %s add %s 0x%x' % (vxlanName, vxlanInfo[2], vxlanInfo[3]))
            self.newVxLanDataEvent.set()

    def IsOperStatePeerCheckNeeded(self, vxlan):
        '''
        In the following cases we don't need to co-ordinate vxlan state bringup
        with the peer -
            1. peer is not alive
            2. peer is alive but doesn't understand the new state machine
            3. vni is not present in the peerDb
            4. peer-vxlan is down on the peer
            5. peer has already dual connected the vxlan - this is a tricky one
               and would result in problem windows when the oper state changes
               on the local vxlan repeatedly and rapidly - XXX need to revist
               once everything becomes event driven.
        '''
        if not self.Intf.isPeerAlive():
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'peer dead; oper-up %s' % (vxlan))
            return False
        
        if not self.Parser.isPeerVxLanStateAware():
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'peer is older version; oper-up %s' % (vxlan))
            return False

        ourInfo = self.GetOurVxLanInfo(vxlan)
        vni = ourInfo[0] if ourInfo else None
        peerInfo = self.GetPeerVxLanInfoByVni(vni) if vni else None
        if not peerInfo:
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'no peer vxlan for vni %d; oper-up %s' % (vni, vxlan))
            return False

        if not self.IsOperReadyForDualConnect(peerInfo[2]):
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'vni %d peer vxlan oper %s; oper-up %s' % (vni, peerInfo[2], vxlan))
            return False

        if (peerInfo[3] & self._VXLAN_DB_FLAGS_DUAL):
            self.Log.log_debug(self.Log._LOG_DEBUG_FSM, 'vni %d peer vxlan already dual-conn 0x%x; oper-up %s' % (vni, peerInfo[3], vxlan))
            return False

        return True

    def IsOperReadyForDualConnect(self, operState):
        return operState in ["up", "dormant", "down"] and \
            self.Intf.isPeerAlive() and self.Parser.isClientSockOpen()

    def CollectLocalInfo(self):
        """
        This function is called on a periodic interval to gather up any 
        information on the local switch which needs to be collected prior to
        sending that information to the peer switch. Because the VxLAN database
        is maintained through the netlink kcache, there is no need to get any
        information in this function.
        """
        pass

    def EncodeXmlInfo(self, parent):
        """
        Encodes the local information as XML.

        parent - An xml.etree.ElementTree.Element instance which is the parent
            element under which all of the information should be encoded.

        <vxlans>
            <vxlan name="vx-14567102" vni="14567102" localip="27.0.0.11" operstate="up" flags="1" op="sync"/>
            <vxlan name="vx-14567103" vni="14567103" localip="27.0.0.11" operstate="up" flags="1" op="sync"/>
            <vxlan name="vx-14567104" vni="14567104" localip="27.0.0.11" operstate="up" flags="1" op="sync"/>
            <vxlan name="vx-14567105" vni="14567105" localip="27.0.0.11" operstate="up" flags="1" op="sync"/>
            <vxlan name="vx-14567106" vni="14567106" localip="27.0.0.11" operstate="up" flags="1" op="sync"/>
        </vxlans>
        """
        vxLanDB = self.GetOurVxLanDB()
        vxlans = ET.SubElement(parent, "vxlans")
        newPeer = self.Parser.isPeerVxLanStateAware()
        for vxlan in vxLanDB:
            vxlanInfo = vxLanDB[vxlan]
            # check if the peer can handle down vxlans
            if not self.IsOperReadyForDualConnect(vxlanInfo[2]) and not newPeer:
                continue
            self.EncodeVxLanXmlInfo(vxlans, vxlan, vxlanInfo, "sync")

    def EncodeVxLanXmlInfo(self, vxlans, vxlan, vxlanInfo, op):
        vxlanAttrib = { "name" : vxlan }
        vxlanAttrib["vni"] = str(vxlanInfo[0])
        vxlanAttrib["localip"] = str(vxlanInfo[1])
        vxlanAttrib["operState"] = str(vxlanInfo[2])
        vxlanAttrib["flags"] = str(vxlanInfo[3])
        vxlanAttrib["op"] = op
        ET.SubElement(vxlans, "vxlan", vxlanAttrib)

    def SendVxLanAdd(self, vxlan):
        if not self.Intf.isPeerAlive() or not self.Parser.isPeerVxLanStateAware():
            return
        vxlanInfo = self.GetOurVxLanInfo(vxlan)
        if vxlanInfo:
            clagDataAttr = {'version' : self.Parser._ClagDataVersion}
            root = ET.Element("clag_data", clagDataAttr)
            vxlans = ET.SubElement(root, "vxlan-updates")
            self.EncodeVxLanXmlInfo(vxlans, vxlan, vxlanInfo, "add")
            xmlStr = ET.tostring(root)
            self.peerSendQueue.put(xmlStr)

    def DecodeVxLanXmlInfo(self, vxlans):
        vxLanDB = {}
        for vxlan in vxlans.findall('vxlan'):
            name = vxlan.get('name')
            vni = int(vxlan.get('vni',0))
            localip = vxlan.get('localip', "0.0.0.0")
            operState = vxlan.get('operState', "up")
            flags = int(vxlan.get('flags', 0))
            op = vxlan.get('op', "sync")
            if name is not None:
                vxLanInfo = ( vni, localip, operState, flags )
                if op == "sync":
                    vxLanDB[name] = vxLanInfo
                elif op == "add":
                    self.AddVxLanToPeerVxLanDB(name, vxLanInfo)
        return vxLanDB

    def DecodeXmlInfo(self, root):
        """
        Decodes XML-encoded information sent from the peer.

        root - The root element of the XML tree
        """
        vxlans = root.find('vxlans')
        if vxlans is not None:
            vxLanDB = self.DecodeVxLanXmlInfo(vxlans)
            self.SetPeerVxLanDB(vxLanDB)

        vxlans = root.find('vxlan-updates')
        if vxlans is not None:
            self.DecodeVxLanXmlInfo(vxlans)

    def DualIfChange(self, dualIfs, enable):
        """
        Called when an interface becomes or is no longer dual connected
        """
        SetOrClearFlags = self.SetVxLanFlags if enable else self.ClearVxLanFlags
        for vxlan in dualIfs:
            SetOrClearFlags(vxlan, self._VXLAN_DB_FLAGS_DUAL)

    def ClearPeerInfo(self):
        """
        Removes information which has been obtained from the peer.
        """
        if self.peerVxLanDBLock.acquire(False):
            if self.peerVxLanDB != {}:
                self.peerVxLanDB = {}
                self.newVxLanDataEvent.set()
            self.peerVxLanDBLock.release()

    def ClearLocalInfo(self):
        """
        Removes information which has been obtained locally.
        """
        self.SetOurVxLanDB({})

    def UpdateVxLanFromPeer(self):
        '''
        This function is an independent thread of the script. It is woken up when there
        are changes to the local or peer vxlanDB. 
        #       "vx-14567103" : ( 14567103, "27.0.0.11", "dormant", 0 )
        '''
        self.log_debug("Beginning execution of the thread UpdateVxLanFromPeer")
        self.currDualIfs = set()

        while self.keep_going:
            # Wait for a change in our bonds or the peer's bonds
            self.log_debug("Waiting for an event before continuing execution of the thread UpdateVxLanFromPeer")
            self.newVxLanDataEvent.wait()
            self.newVxLanDataEvent.clear()
            self.log_debug("Executing the thread UpdateVxLanFromPeer.")

            # Are we shutting down?
            if not self.keep_going:
                break

            # Is the VxLAN anycast IP configured?
            if not self.Parser.args.vxlanAnycast:
                clagIntfs = [ intf for intf in self.Parser.GetClagIntfDB().values() \
                    if intf.intfType == self.Parser._CLAG_INTF_TYPE_VXLAN ]
                params = []
                self.Intf.protoDownLock.acquire()
                for intf in clagIntfs:
                    self.log_debug("No anycast IP, setting proto down for interface %s " % (intf.intfName,))
                    params += intf.UpdateProtoDownFlags(intf._PROTO_DOWNF_NO_ANYCAST, 0, batch=True)
                self.Intf.ExecuteBatch(["/sbin/ip", "-force", "-batch"], "link set dev %s protodown %s", params)
                self.Intf.protoDownLock.release()
                continue

            # Has the role changed? If so, refresh the interface programming.
            if self.newRole:
                self.newRole = False
                self.RefreshDualIfs()

            # Sets are used to determine which interfaces are newly or no longer dual conencted
            # Filter out vxlans that are not eligible for dual connection
            vxLanDB = self.GetOurVxLanDB()
            ourVxLanDB = {k:vxLanDB[k] for k in vxLanDB if self.IsOperReadyForDualConnect(vxLanDB[k][2])}
            ourVniDB = {v[0]:k for k,v in ourVxLanDB.iteritems()}

            # Filter out bonds in the peerDB that are not eligible for dual 
            # connection
            peerDB = self.GetPeerVxLanDB()
            peerVxLanDB = {k:peerDB[k] for k in peerDB if self.IsOperReadyForDualConnect(peerDB[k][2])}
            peerVniDB = {v[0]:k for k,v in peerVxLanDB.iteritems()}

            dualVnis = set(ourVniDB) & set(peerVniDB)
            newDualIfs = { ourVniDB[vni] for vni in dualVnis }

            nowDualIfs = newDualIfs - self.currDualIfs
            notDualIfs = self.currDualIfs - newDualIfs

            self.log_debug("ourVnis     : %s" % (str(ourVniDB),))
            self.log_debug("peerVnis    : %s" % (str(peerVniDB),))
            self.log_debug("newDualIfs  : %s" % (newDualIfs,))
            self.log_debug("currDualIfs : %s" % (self.currDualIfs,))
            self.log_debug("nowDualIfs  : %s" % (nowDualIfs,))
            self.log_debug("notDualIfs  : %s" % (notDualIfs,))

            # Handle the new dually connected interfaces
            if nowDualIfs:
                self.DualHomeIf(newDualIfs, nowDualIfs, True)

            # Handle the interfaces which are no longer dual connected
            if notDualIfs:
                self.Intf.DelDualIfs(notDualIfs)
                self.DualHomeIf(newDualIfs, notDualIfs, False)

            # Update the current notion of dual connected VxLANs in the global database.
            theDualIfs = {}
            for vni in dualVnis:
                theDualIfs[ourVniDB[vni]] = peerVniDB[vni]
            self.Intf.AddDualIfs(theDualIfs)

            for notDualIf in notDualIfs:
                self.Log.log("%s is no longer dual connected." % (notDualIf,))

            for nowDualIf in nowDualIfs:
                self.Log.log("%s is now dual connected." % (nowDualIf,))

            for sync in self.DataSyncs:
                if not self.keep_going:
                    break
                sync.DualIfChange(nowDualIfs, True)

            for sync in self.DataSyncs:
                if not self.keep_going:
                    break
                sync.DualIfChange(notDualIfs, False)

            self.currDualIfs = newDualIfs

        # When leaving, revert the local IP address to original value and remove
        # anycast IP from loopback inteface.
        DEVNULL = open(os.devnull, 'wb')
        anycastIp = self.Parser.args.vxlanAnycast
        if anycastIp:
            cmd = [ "/sbin/ip", "addr", "del", "dev", "lo", str(anycastIp) + "/32", "scope", "global" ]
            self.log_debug("Executing: %s " % (cmd,))
            subprocess.call(cmd, stdout=DEVNULL, stderr=DEVNULL)

        vxlanDB = self.GetOurVxLanDB()
        ipParams = []
        bridgeParams = []
        for (vxlanName, vxlanData) in vxlanDB.iteritems():
            # Reset the local IP for the VxLAN if not on the primary node
            if vxlanData[1]:
                ipParams += [(vxlanName, vxlanData[1])]
            bridgeParams += [(vxlanName)]
        self.Intf.ExecuteBatch(["/sbin/ip", "-force", "-batch"], "link set dev %s type vxlan local %s", ipParams)
        self.Intf.bridgeAttrLock.acquire()
        self.Intf.ExecuteBatch(["/sbin/bridge", "-force", "-batch"], "link set dev %s duallink off", bridgeParams)
        self.Intf.bridgeAttrLock.release()

        self.log_debug("Finished execution of the thread UpdateVxLanFromPeer")

    def DualHomeIfGetIntfs(self, dualIfs):
        # Create a set of all peer sub interfaces 
        masterVlanMap = self.Intf.GetMasterIfVlanMap()
        peerMasters = set([self.Parser.args.peerIf])
        if self.Intf.isSubIfName(self.Parser.args.peerIf):
            peerMasters = set(self.Intf.GetLogicalMastersOfSubIf(self.Parser.args.peerIf))
        peerSet = peerMasters.copy()
        for peerMaster in peerMasters:
            peerSet |= set(masterVlanMap.get(peerMaster, {}).values())

        # Create a set of all dual connected sub interfaces
        dualSet = set(dualIfs)
        for dualIf in dualIfs:
            dualSet |= set(masterVlanMap.get(dualIf, {}).values())
        return (peerSet, dualSet)

    def SetSwitchdAnycastFuseNode(self):
        '''
        This function should go away eventually, since sideband access to switchd
        is a bad thing. But for now, this write the anycastIp and the localIp to
        the switchd /config/vxlan/anycast_ip fuse node.
        '''
        anycastIp = self.Parser.args.vxlanAnycast
        if anycastIp:
            localIp = "0.0.0.0"
            vxlanDB = self.GetOurVxLanDB()
            if vxlanDB:
                localIp = vxlanDB.itervalues().next()[1]
            try:
                with open("/cumulus/switchd/config/vxlan/local_ip", "w") as f:
                    f.write(str(anycastIp) + " " + str(localIp))
            except IOError:
                pass

    def DualHomeIf(self, allDualIfs, dualIfs, enable):
        '''
        Enable or disable the forwarding logic for a dual homed interface.
        '''
        # Are interfaces becoming dual connected?
        (peerSet, dualSet) = self.DualHomeIfGetIntfs(dualIfs)
        DEVNULL = open(os.devnull, 'wb')
        if enable:
            anycastIp = self.Parser.args.vxlanAnycast
            if anycastIp:
                # Is this the first set of dual-connected VxLANs?
                if allDualIfs == dualIfs:
                    # XXX: Remove this eventually. Side-band access to switchd to inform of the anycast IP
                    self.SetSwitchdAnycastFuseNode()

                clagIntfs = [ intf for intf in self.Parser.GetClagIntfDB().values() \
                    if intf.intfType == self.Parser._CLAG_INTF_TYPE_VXLAN ]
                params = []
                for intf in clagIntfs:
                    ipaddr = None
                    if intf.intfName in dualIfs:
                        ipaddr = anycastIp
                    elif allDualIfs == dualIfs and allDualIfs:
                        if self.Parser.clagRole == self.Parser._ROLE_PRIMARY:
                            ourVxlan = self.GetOurVxLanInfo(intf.intfName)
                            ipaddr = ourVxlan[1] if ourVxlan else None
                    if ipaddr:
                        params += [(intf.intfName, str(ipaddr))]
                self.Intf.ExecuteBatch(["/sbin/ip", "-force", "-batch"], "link set dev %s type vxlan local %s", params)

                # For each bridge link, set the duallink property
                params = []
                for dualIfName in dualSet:
                    params += [(dualIfName,)]
                self.Intf.bridgeAttrLock.acquire()
                self.Intf.ExecuteBatch(["/sbin/bridge", "-force", "-batch"], "link set dev %s duallink on", params)
                self.Intf.bridgeAttrLock.release()

                # Clear proto-down for the newly dual connected interfaces, and if these are the 
                # first dual-connected interfaces then all other (singly-connected) VxLANs go proto down
                params = []
                self.Intf.protoDownLock.acquire()
                for intf in clagIntfs:
                    if intf.intfName in dualIfs:
                        self.log_debug("Dual Connect, clearing proto down for interface %s " % (intf.intfName,))
                        params += intf.UpdateProtoDownFlags(0, intf._PROTO_DOWNF_SINGLE_VXLAN, batch=True)
                    elif allDualIfs == dualIfs and allDualIfs:
                        self.log_debug("Dual Connect, setting proto down for interface %s " % (intf.intfName,))
                        params += intf.UpdateProtoDownFlags(intf._PROTO_DOWNF_SINGLE_VXLAN, 0, batch=True)
                self.Intf.ExecuteBatch(["/sbin/ip", "-force", "-batch"], "link set dev %s protodown %s", params)
                self.Intf.protoDownLock.release()

                # When the first interface comes up, program the anycast IP on the loopback interface
                if allDualIfs == dualIfs and allDualIfs:
                    self.Intf.NetlinkFilter({socket.AF_BRIDGE}, set())
                    self.log_debug("Waiting for these interfaces to go forwarding: %s " % (str(dualIfs),))
                    self.waitForUp = set(dualIfs)
        else:
            # Interfaces are becoming singly connected
            # Is this the last set of dual-connected VxLANs
            anycastIp = self.Parser.args.vxlanAnycast
            if not allDualIfs:
                # Remove the anycast IP on the loopback interface except on the primary
                if anycastIp and self.Parser.clagRole != self.Parser._ROLE_PRIMARY:
                    cmd = [ "/sbin/ip", "addr", "del", "dev", "lo", str(anycastIp) + "/32", "scope", "global" ]
                    self.log_debug("Executing: %s " % (cmd,))
                    subprocess.call(cmd, stdout=DEVNULL, stderr=DEVNULL)
                    self.Intf.NetlinkFilter(set(), {socket.AF_BRIDGE})
                    self.waitForUp = set()

            # Clear the VxLAN's proto-down flags
            clagIntfs = [ intf for intf in self.Parser.GetClagIntfDB().values() \
                if intf.intfType == self.Parser._CLAG_INTF_TYPE_VXLAN ]
            params = []
            self.Intf.protoDownLock.acquire()
            for intf in clagIntfs:
                if not allDualIfs:
                    if self.Parser.clagRole == self.Parser._ROLE_PRIMARY:
                        params += intf.UpdateProtoDownFlags(0, intf._PROTO_DOWNF_SINGLE_VXLAN, batch=True)
                    else:
                        params += intf.UpdateProtoDownFlags(intf._PROTO_DOWNF_SINGLE_VXLAN, 0, batch=True)
                elif intf.intfName in dualIfs:
                    params += intf.UpdateProtoDownFlags(intf._PROTO_DOWNF_SINGLE_VXLAN, 0, batch=True)
            self.Intf.ExecuteBatch(["/sbin/ip", "-force", "-batch"], "link set dev %s protodown %s", params)
            self.Intf.protoDownLock.release()

            # For each bridge link, reset the duallink property
            params = []
            for dualIfName in dualSet:
                params += [(dualIfName,)]
            self.Intf.bridgeAttrLock.acquire()
            self.Intf.ExecuteBatch(["/sbin/bridge", "-force", "-batch"], "link set dev %s duallink off", params)
            self.Intf.bridgeAttrLock.release()

            params = []
            for intf in clagIntfs:
                if not self.keep_going:
                    break
                ourVxlan = self.GetOurVxLanInfo(intf.intfName)
                localIp = ourVxlan[1] if ourVxlan else None
                ipaddr = None
                if not allDualIfs:
                    if self.Parser.clagRole == self.Parser._ROLE_PRIMARY:
                        ipaddr = anycastIp
                    else:
                        ipaddr = localIp
                elif intf.intfName in dualIfs:
                    ipaddr = localIp
                if ipaddr:
                    params += [(intf.intfName, str(ipaddr))]
            self.Intf.ExecuteBatch(["/sbin/ip", "-force", "-batch"], "link set dev %s type vxlan local %s", params)

    def RefreshDualIfs(self):
        '''
        Refresh the programming of the dual connected vxlan devices, and those
        which are not dual connected.
        '''
        clagIntfs = set([ intf.intfName for intf in self.Parser.GetClagIntfDB().values() \
            if intf.intfType == self.Parser._CLAG_INTF_TYPE_VXLAN ])
        notDualIfs = clagIntfs - self.currDualIfs
        self.DualHomeIf(self.currDualIfs, self.currDualIfs, True)
        self.DualHomeIf(self.currDualIfs, notDualIfs, False)

    def UpdateVxLanConfig(self):
        '''
        This function is an independent thread of the script. It periodically 
        runs to check if the learning priority and learning filter are properly
        set. It is possible that due to links going up and down outside of the
        control of CLAG, these settings can be modified.
        '''
        self.log_debug("Beginning execution of the thread UpdateVxLanConfig")
        runCnt = 0
        currRole = None

        while self.keep_going:

            # Look for a role change
            if currRole != self.Parser.clagRole:
                currRole = self.Parser.clagRole
                self.log_debug("Role has changed to %s, refreshing interface states" % ("Primary" if currRole == self.Parser._ROLE_PRIMARY else "Secondary",))
                self.newRole = True
                self.newVxLanDataEvent.set()

            self.log_debug("Waiting for an event before continuing execution of the thread UpdateVxLanConfig")
            self.stopEvent.wait(self.Parser.getPeriodicRun())
            self.log_debug("Executing the thread UpdateVxLanConfig")

        self.log_debug("Finished execution of the thread UpdateVxLanConfig")

